# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import copy, deepcopy
from enum import Enum
from math import isclose
from typing import Dict, List, Tuple, Union, cast

import matplotlib.pyplot as plt
import numpy as np
from numpy.polynomial.polynomial import Polynomial
from pytket import Circuit, OpType, Qubit
from pytket.backends import Backend
from pytket.circuit import Node
from pytket.pauli import Pauli, QubitPauliString
from pytket.predicates import CompilationUnit
from pytket.unit_id import UnitID, UnitType
from pytket.utils import QubitPauliOperator
from pytket.utils.operators import CoeffTypeAccepted
from scipy.optimize import curve_fit  # type: ignore
from sympy import Expr  # type: ignore

from qermit import (
    AnsatzCircuit,
    MitEx,
    MitRes,
    MitTask,
    ObservableExperiment,
    ObservableTracker,
    TaskGraph,
)
from qermit.noise_model import NoiseModel, PauliErrorTranspile

box_types = {
    OpType.CircBox,
    OpType.ExpBox,
    OpType.PauliExpBox,
    OpType.Unitary1qBox,
    OpType.Unitary2qBox,
    OpType.Unitary3qBox,
}


class Folding(Enum):
    """Folding techniques used to increase the noise levels.

    :return: Circuit with gate count increased to scale noise. The unitary implemented by the
        circuit has not changed.
    """

    # TODO: circ does not appear as input in docs
    # TODO Generalise with 'partial folding' to allow for non integer noise scaling
    @staticmethod
    def circuit(circ: Circuit, noise_scaling: int, **kwargs) -> List[Circuit]:
        """Noise scaling by circuit folding. In this case the folded circuit is of
        the form :math:`CC^{-1}CC^{-1}...C` where :math:`C` is the original circuit. As such noise may be scaled by
        odd integers. The Unitary implemented is unchanged by this process.

        :param circ: Original circuit to be folded.
        :param noise_scaling: Factor by which to scale the noise. This must be an odd integer.
        :raises ValueError: Raised if the amount by which the noise should be scaled is not an odd integer.
        :return: Folded circuit implementing identical unitary to the initial circuit.
        """

        # Raise if the amount by which the noise should be scaled is not an odd integer
        if (not noise_scaling % 2) or noise_scaling % 1:
            raise ValueError(
                "When performing circuit folding, the noise scaling must be an odd integer."
            )

        folded_circ = circ.copy()
        for _ in range(noise_scaling // 2):
            # Add barrier between circuit and its inverse
            folded_circ.add_barrier(
                cast(List[UnitID], folded_circ.qubits + folded_circ.bits)
            )

            # Add inverse circuit by iterating though commands and inverting them
            for gate in reversed(circ.get_commands()):
                if gate.op.type == OpType.Barrier:
                    folded_circ.add_barrier(gate.args)
                elif gate.op.type in box_types:
                    raise RuntimeError("Box types not supported when folding.")
                else:
                    folded_circ.add_gate(gate.op.dagger, gate.args)

            # Add barrier between circuit and its inverse
            folded_circ.add_barrier(
                cast(List[UnitID], folded_circ.qubits + folded_circ.bits)
            )

            # Add original circuit
            for gate in circ.get_commands():
                if gate.op.type == OpType.Barrier:
                    folded_circ.add_barrier(gate.args)
                elif gate.op.type in box_types:
                    raise RuntimeError("Box types not supported when folding.")
                else:
                    folded_circ.add_gate(gate.op, gate.args)

        return [folded_circ]

    @staticmethod
    def two_qubit_gate(circ: Circuit, noise_scaling: float, **kwargs) -> List[Circuit]:
        """Noise scaling by folding 2 qubit gates. It is implicitly
        assumed that the noise on the 2 qubit gates dominate. Two qubit gates
        :math:`G` are replaced by :math:`GG^{-1}G...G^{-1}G`. If
        `noise_scaling` is of the form (#gates + 2i)/#gates,
        where #gates is the number of gates in the compiled circuit and i is
        an integer, then the noise scaling is exact. It will otherwise be
        as close as possible to but smaller then noise_scaling.

        :param circ: Original circuit to be folded.
        :param noise_scaling: Factor by which the noise should be scaled.

        :raises ValueError: Raised if noise_scaling is less than 1.
        :raises ValueError: Raised if the noise cannot be scaled by
            exactly noise_scaling and `_allow_approx_fold` is not True.
        :raises RuntimeError: Raised if there are no valid gates to fold.
        :raises RuntimeError: Raised if the circuit includes boxes.

        :key _allow_approx_fold: True or false depending on if
            approximate folding is allowed. Defaults to True.

        :return: Circuit with noise scaled.
        """

        if noise_scaling < 1:
            raise ValueError("noise_scaling must be greater than or equal to 1")

        _allow_approx_fold = kwargs.get("_allow_approx_fold", True)

        # All gates will be folded by an amount equal to the even number
        # less than noise_scaling-1. By doing so the noise is scaled by the
        # odd integer less than noise_scaling.
        num_folds_dict = {
            i: (int(noise_scaling - 1) // 2)
            for i, cmd in enumerate(circ.get_commands())
            if (
                not (cmd.op.type == OpType.Barrier)
                and (cmd.op.type not in box_types)
                and (len(cmd.qubits) == 2)
            )
        }

        if len(num_folds_dict) == 0:
            raise RuntimeError(
                "There are no valid q qubits gates in this circuit to fold. "
                "Your circuit should include 2 qubit gates other than "
                "Barrier and CircBox."
            )

        # The remaining noise scaling is achieved by randomly selecting gates
        # to scale. fold_frac gives the fraction of gates which need to be
        # folded to achieve noise_scaling. The appropriate fraction of
        # gates is then randomly selected.
        fold_frac = ((noise_scaling - 1) % 2) / 2
        to_fold = np.random.choice(
            list(num_folds_dict.keys()),
            size=int(len(num_folds_dict.keys()) * fold_frac),
            replace=False,
        )
        for i in to_fold:
            num_folds_dict[i] += 1

        true_noise_scaling = float(sum(2 * i + 1 for i in num_folds_dict.values()))
        true_noise_scaling /= len(num_folds_dict)

        if not (
            _allow_approx_fold
            or isclose(noise_scaling, true_noise_scaling, abs_tol=0.001)
        ):
            raise ValueError(
                "The noise cannot be scaled by the amount inputted."
                "The noise must be scaled by a factor of the form "
                "(#gates + 2i)/#gates, where #gates is the number of gates "
                "in the compiled circuit, and i is an integer."
            )

        # Copy qubit register of original circuit.
        folded_circuit = Circuit()
        for qubit in circ.qubits:
            folded_circuit.add_qubit(qubit)

        for i, gate in enumerate(circ.get_commands()):
            # Barriers are not folded and added as given.
            if gate.op.type == OpType.Barrier:
                folded_circuit.add_barrier(gate.args)
            # Boxes are not supported.
            elif gate.op.type in box_types:
                raise RuntimeError("Box types not supported when folding.")
            # 2 qubit gates are folded.
            elif len(gate.qubits) == 2:
                folded_circuit.add_gate(gate.op, gate.args)
                for _ in range(num_folds_dict[i]):
                    folded_circuit.add_barrier(gate.args)
                    folded_circuit.add_gate(gate.op.dagger, gate.args)
                    folded_circuit.add_barrier(gate.args)
                    folded_circuit.add_gate(gate.op, gate.args)
            # All other gates are added as given.
            else:
                folded_circuit.add_gate(gate.op, gate.args)

        return [folded_circuit]

    @staticmethod
    def gate(circ: Circuit, noise_scaling: float, **kwargs) -> List[Circuit]:
        """Noise scaling by gate folding. In this case gates :math:`G` are replaced at random
        with :math:`GG^{-1}G` until the number of gates is sufficiently scaled.

        :param circ: Original circuit to be folded.
        :param noise_scaling: Factor by which to increase the noise.

        :key _allow_approx_fold: Allows for the noise to be increased by an amount close to that requested, as
            opposed to by exactly the amount requested.
            This is necessary as there are cases where the exact noise scaling cannot be achieved.
            This occurs due to the discrete
            amounts by which the noise can be increased (i.e. the discrete amount by which one gate increases the noise).

        :raises ValueError: Raised if the requested noise scaling cannot be exactly achieved. This can be
            avoided by appropriately setting _allow_approx_fold.
        :return: Folded circuit implementing identical unitary to the initial circuit.
        """

        _allow_approx_fold = kwargs.get("_allow_approx_fold", 0)

        c_dict = circ.to_dict()
        num_commands = len(c_dict["commands"])

        # Count and locate the number of commands that are not barriers
        c_dict_commands_non_barrier = [
            (i, command)
            for (i, command) in enumerate(c_dict["commands"])
            if command["op"]["type"] != "Barrier"
        ]
        num_non_barrier_commands = len(c_dict_commands_non_barrier)

        # Calculate the number of gates that need to be folded. Note that only
        # non barrier commands should be folded so only these are counted.
        num_additional_commands = (noise_scaling - 1) * num_non_barrier_commands
        num_folded_commands = int(num_additional_commands // 2)

        true_noise_scaling = 1 + ((num_folded_commands * 2) / num_non_barrier_commands)
        # This check isolates the case where the noise folding that can be achieved is different
        # from that requested. While noise_scaling can be any real value, as the gates increase
        # the noise by discrete values, not all folding values are possible.
        if (not _allow_approx_fold) and (
            abs(noise_scaling - true_noise_scaling) > 0.001
        ):
            raise ValueError(
                "The noise cannot be scaled by the amount inputted. The noise must be scaled by a factor of the form (#gates + 2i)/#gates, where #gates is the number of gates in the compiled circuit, and i is an integer."
            )

        # Choose a random selection of commands to fold. Commands are referenced by their index.
        # These are chosen with replacement. Only non barrier commands should
        # be folded, and so only the index of those are added here.
        commands_to_fold = np.random.choice(
            [i[0] for i in c_dict_commands_non_barrier], num_folded_commands
        )
        # Calculate how many times each individual command needs to be folded
        num_folds = {i: list(commands_to_fold).count(i) for i in range(num_commands)}

        command_circ_dict = {
            key: val for key, val in c_dict.items() if key != "commands"
        }

        folded_command_list = []
        # For each command, fold the appropriate number of times.
        for command_index in num_folds:
            command = c_dict["commands"][command_index]
            command_circ_dict.update({"commands": [command]})
            command_circ = Circuit().from_dict(command_circ_dict)

            # Commands which are not to be folded may not be invertible (for
            # example barriers) and so the inverse is not calculated in that
            # case.
            if num_folds[command_index] > 0:
                # Find the inverse of the command
                inverse_command_circ = command_circ.dagger()
                inverse_command_circ_dict = inverse_command_circ.to_dict()
                inverse_command = inverse_command_circ_dict["commands"]

            # Append command and inverse the appropriate number of times.
            folded_command_list.append(command)
            for _ in range(num_folds[command_index]):
                folded_command_list.append(
                    {
                        "args": command["args"],
                        "op": {
                            "signature": ["Q"] * len(command["args"]),
                            "type": "Barrier",
                        },
                    }
                )
                folded_command_list.append(*inverse_command)
                folded_command_list.append(
                    {
                        "args": command["args"],
                        "op": {
                            "signature": ["Q"] * len(command["args"]),
                            "type": "Barrier",
                        },
                    }
                )
                folded_command_list.append(command)

        folded_c_dict = c_dict.copy()
        folded_c_dict["commands"] = folded_command_list
        folded_c = Circuit().from_dict(folded_c_dict)

        return [folded_c]

    @staticmethod
    def odd_gate(circ: Circuit, noise_scaling: int, **kwargs) -> List[Circuit]:
        """Noise scaling by gate folding. In this case odd gates :math:`G` are
        replaced :math:`GG^{-1}G` until the number of gates is sufficiently
        scaled.

        :param circ: Original circuit to be folded.
        :param noise_scaling: Factor by which to increase the noise.
        :return: Folded circuit implementing identical unitary to the initial circuit.
        """

        c_dict = circ.to_dict()

        fold = True
        folded_command_list = []

        command_circ_dict = {
            key: val for key, val in c_dict.items() if key != "commands"
        }

        for command in c_dict["commands"]:
            # Barriers are added to the circuit but otherwise effectively
            # skipped.
            if command["op"]["type"] == "Barrier":
                folded_command_list.append(command)

            elif fold:
                command_circ_dict.update({"commands": [command]})
                command_circ = Circuit.from_dict(command_circ_dict)

                # Find the inverse of the command
                inverse_command_circ = command_circ.dagger()
                inverse_command_circ_dict = inverse_command_circ.to_dict()
                inverse_command = inverse_command_circ_dict["commands"]

                # Append command and inverse the appropriate number of times.
                folded_command_list.append(command)
                for _ in range(noise_scaling - 1):
                    folded_command_list.append(
                        {
                            "args": command["args"],
                            "op": {
                                "signature": ["Q"] * len(command["args"]),
                                "type": "Barrier",
                            },
                        }
                    )
                    folded_command_list.append(*inverse_command)
                    folded_command_list.append(
                        {
                            "args": command["args"],
                            "op": {
                                "signature": ["Q"] * len(command["args"]),
                                "type": "Barrier",
                            },
                        }
                    )
                    folded_command_list.append(command)

                fold = not fold
            else:
                folded_command_list.append(command)

                fold = not fold

        c_dict.update({"commands": folded_command_list})
        folded_c = Circuit.from_dict(c_dict)

        return [folded_c]

    @staticmethod
    def noise_aware(circ: Circuit, noise_scaling: int, **kwargs) -> List[Circuit]:
        """Scale noise in a circuit by adding noisy gates as defined
        by the given noise model.

        :param circ: Circuit with noise to be scaled.
        :param noise_scaling: Factor by which noise should be scaled.
        :return: List of circuits with additional noise gates added.

        :key noise_model: Noise model defining noise types and rates.
            Defaults to noiseless model.
        :key n_noisy_circuit_samples: The number of random noisy
            circuits to generate. Defaults to 1.
        """

        noise_model: NoiseModel = kwargs.get("noise_model", NoiseModel(noise_model={}))
        n_noisy_circuit_samples: int = kwargs.get("n_noisy_circuit_samples", 1)

        scaled_noise_model: NoiseModel = noise_model.scale(
            scaling_factor=noise_scaling - 1
        )
        error_transpiler = PauliErrorTranspile(noise_model=scaled_noise_model)

        scaled_circ_list = []
        for _ in range(n_noisy_circuit_samples):
            scaled_circ = circ.copy()
            error_transpiler.apply(scaled_circ)
            scaled_circ_list.append(scaled_circ)

        return scaled_circ_list


def poly_exp_func(x: float, *params) -> float:
    """Definition of poly-exponential function for the purposes of fitting to data

    :param x: Value at which to evaluate function
    :return: Evaluation of poly-exponential function
    """
    # Note that we use a list ending in 0 so that the constant in the polynomial is 0.
    # The constant can then be absorbed into the coefficient of the exponential.
    return params[0] + params[1] * np.exp(np.polyval([*params[2:], 0], x))


def cube_root_func(x: float, *params) -> float:
    """Definition of cube root function for the purposes of fitting to data.

    :param x: Value at which to evaluate cube root function
    :return: Evaluation of cube root function
    """
    y = x + params[2]
    return params[0] + params[1] * np.sign(y) * (np.abs(y)) ** (1 / 3)


class Fit(Enum):
    """Functions to fit to expectation values as they change with noise.

    :return: Extrapolation of expectation values to the zero noise limit.
    """

    # TODO Consider adding adaptive exponential extrapolation
    @staticmethod
    def cube_root(x: List[float], y: List[float], _show_fit: bool, *args) -> float:
        """Fit data to a cube root function. This is to say a function of the form :math:`a + b(x+c)^{1/3}`.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitted function.
        :return: Extrapolation of data to zero noise limit using the best fitting cube root function.
        """

        # Fit data to cube root function
        vals = curve_fit(cube_root_func, x, y, p0=[0, 1, 0], maxfev=100000)

        # Evaluate fitted function at 0
        fit_to_zero = cube_root_func(0, *vals[0])

        # Plot fitted function and data
        if _show_fit:
            fit_x = np.linspace(0, max(x), 100)
            fit_y = [cube_root_func(i, *vals[0]) for i in fit_x]

            plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)

        return float(fit_to_zero)

    @staticmethod
    def poly_exponential(
        x: List[float], y: List[float], _show_fit: bool, deg: int
    ) -> float:
        """Fit data to a poly-exponential, which is to say a function of the
        form :math:`a+e^{z}`, where :math:`z` is a polynomial.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitted function.
        :param deg: The degree of the polynomial in the exponential.
        :raises ValueError: Raised if the degree of the polynomial
            inputted is negative, or too high to fit to the data.
        :return: Extrapolation of data to the zero noise limit using the best fitting
            poly-exponential function of the specified degree.
        """

        # check that the degree of the polynomial is positive, and small enough to fit
        # to the inputted data
        if (deg + 2 > len(x)) or (deg < 0):
            raise ValueError(
                "The degree of the polynomial must be positive and can be at most is m-2, where m is the number of data points. In this case the largest permitted degree is %i."
                % (len(x) - 2)
            )

        # Fit data to polyexponential function
        # TODO: Improve bounds here.
        bounds = (
            [-1, -2, *[-np.inf for i in range(deg)]],
            [1, 2, *[np.inf for i in range(deg)]],
        )

        # Initialise decaying poly-exponential with intersect at
        # unfolded noisy value.
        least_noisy_y_index = x.index(1)
        p0 = [0, y[least_noisy_y_index], *[-1 for i in range(deg)]]

        vals = curve_fit(
            poly_exp_func,
            x,
            y,
            p0=p0,
            maxfev=10000,
            bounds=bounds,
        )

        # Extrapolate function to zero noise limit
        fit_to_zero = poly_exp_func(0, *vals[0])

        # Plot data and fitted function
        if _show_fit:
            fit_x = np.linspace(0, max(x), 100)
            fit_y = [poly_exp_func(i, *vals[0]) for i in fit_x]

            plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)

        return float(fit_to_zero)

    @staticmethod
    def exponential(x: List[float], y: List[float], _show_fit: bool, *args) -> float:
        """Fit data to an exponential function. This is to say a function of the form :math:`a+e^{(b+x)}`.
        Note that this is a special case of the poly-exponential function.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitting function.
        :return: Extrapolation to zero noise limit using the best fitting exponential function.
        """

        # As the exponential function is a special case of the
        # poly-exponential function, it is called here
        return Fit.poly_exponential(x, y, _show_fit, 1)

    @staticmethod
    def polynomial(x: List[float], y: List[float], _show_fit: bool, deg: int) -> float:
        """Fit data to a polynomial function.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitting function.
        :param deg: The degree of the function to fit to.
        :raises ValueError: Raised if the degree of the polynomial is negative,
            or too high to fit the data to.
        :return: Extrapolation to zero noise limit using the best fitting polynomial
            function of the specified degree.
        """

        # Raised if the degree of the polynomial requested is negative, or too high to fit the data to
        if (deg > len(x)) or (deg < 0):
            raise ValueError(
                "The degree of the polynomial must be positive and can be at most m-1, where m is the number of data points."
            )

        # Fit data to a polynomial
        fit = Polynomial.fit(x, [float(i) for i in y], deg=deg)

        # Extrapolate to the zero noise limit
        fit_to_zero = fit.convert().coef[0]

        # Plot data and fitted function
        if _show_fit:
            linspace = fit.linspace()
            fit_x = linspace[0]
            fit_y = linspace[1]

            plot_fit(x, y, list(fit_x), list(fit_y), fit_to_zero)

        return float(fit_to_zero)

    @staticmethod
    def linear(x: List[float], y: List[float], _show_fit: bool, *args) -> float:
        """Fit data to a linear function. This is to say a function of the form :math:`ax+b`.
        Note that this is a special case of the polynomial fitting function.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitted function.
        :return: Extrapolation to zero noise limit using the best fitting linear function.
        """
        # As this is a special case of a fit to a polynomial, the polynomial
        # fitting function is called here with a degree 1
        return Fit.polynomial(x, y, _show_fit, 1)

    @staticmethod
    def richardson(x: List[float], y: List[float], _show_fit: bool, *args) -> float:
        """Use richardson extrapolation. This amounts to fitting to a polynomial of
        degree one less than the number of data points.

        :param x: Noise scaling values.
        :param y: Expectation values.
        :param _show_fit: Plot data and resulting fitted function.
        :return: Extrapolation to zero noise limit using Richardson extrapolation.
        """
        # As this is a special case of the polynomial fitting function, the polynomial fitting
        # function is called here with degree one less than the number of data points.
        return Fit.polynomial(x, y, _show_fit, len(x) - 1)


def plot_fit(
    x: List[float],
    y: List[float],
    fit_x: List[float],
    fit_y: List[float],
    fit_to_zero: float,
):
    """Plot expectation values at each noise level, and the fit to the data derived.

    :param x: Amounts by which the noise has been scaled
    :param y: Expectation values at each noise level
    :param fit_x: x coordinates at which to plot value of fitted function
    :param fit_y: Value of fitted function at each noise scaling
    :param fit_to_zero: The extrapolation of the fitted function to the zero noise limit
    """

    fig = plt.figure()
    ax1 = fig.add_subplot(111)

    ax1.scatter(x, y, s=10, c="b", marker="s", label="data")
    ax1.plot(fit_x, fit_y, c="g", label="fit")

    ax1.scatter(0, fit_to_zero, s=10, c="r", marker="o", label="prediction")

    plt.legend()
    plt.show()


def digital_folding_task_gen(
    backend: Backend,
    noise_scaling: float,
    _folding_type: Folding,
    _allow_approx_fold: bool,
    **kwargs,
) -> MitTask:
    """
    Generates task transforming a circuit in order to amplify the noise. The noise
    is increased by a factor noise_scaling using the inputted folding method.

    :param backend: This will be used to compile the circuit after folding to ensure
        that the gate set matches those available on the backend.
    :param noise_scaling: The factor by which the noise is increased.
    :param _folding_type: The means by which the noise should be increased.
    :param _allow_approx_fold:  Allows for the noise to be increased by an amount close to that requested, as
        opposed to by exactly the amount requested.
        This is necessary as there are cases where the exact noise scaling cannot be achieved.
        This occurs due to the discrete
        amounts by which the noise can be increased (i.e. the discrete amount by which one gate increases the noise).
    """

    def task(
        obj,
        mitex_wire: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment], List[int]]:
        """Increase the noise levels impacting the circuit by increasing the
        number of gates. This preserves the action of the circuit.

        :param mitex_wire: List of experiments
        :return: List of equivalent circuits, but with noise levels increased.
            Each noise scaling value may generate multiple circuits.
            As such the return includes a lit of integers indicating to which
            of the original circuits the new circuits belong.
        """

        folded_circuits = []
        experiment_index = []

        # For each circuit in the input wire, extract the circuit, apply the fold,
        # and perform the necessary compilation.
        for index, experiment in enumerate(mitex_wire):
            # Apply the necessary folding method
            zne_circ_list = _folding_type(
                experiment.AnsatzCircuit.Circuit,
                noise_scaling,
                _allow_approx_fold=_allow_approx_fold,
                **kwargs,
            )  # type: ignore

            for zne_circ in zne_circ_list:
                # TODO: This additional compilation pass may result in the circuit noise being
                # increased too much, and should be removed or better accounted for.

                # This compilation pass was added to account for the case that
                # the inverse of a gate is not in the gateset of the backend.
                backend.rebase_pass().apply(zne_circ)

                folded_circuits.append(
                    ObservableExperiment(
                        AnsatzCircuit=AnsatzCircuit(
                            Circuit=zne_circ,
                            Shots=experiment.AnsatzCircuit.Shots // len(zne_circ_list),
                            SymbolsDict=deepcopy(experiment.AnsatzCircuit.SymbolsDict),
                        ),
                        ObservableTracker=deepcopy(experiment.ObservableTracker),
                    )
                )
                experiment_index.append(index)

        return (folded_circuits, experiment_index)

    return MitTask(_label="DigitalFolding", _n_in_wires=1, _n_out_wires=2, _method=task)


def merge_experiments_task_gen() -> MitTask:
    """Generates task merging qubit pauli strings when they belong to the
    same experiment.

    :return: MitTask performing the merge.
    """

    def task(
        obj, qpo_list: List[QubitPauliOperator], experiment_index_list: List[int]
    ) -> Tuple[List[QubitPauliOperator]]:
        """Merge a list of qubit pauli operators. Qubit pauli operators will
        merged if they belong to the same experiment. The experiment each
        qubit pauli operator belongs to is indicated by the index list.
        Merging here means that the mean of the pauli string coefficients
        is taken to be the coefficient of the qubit pauli string in the
        new qubit pauli operator.

        :param qpo_list: List of qubit pauli strings, some of which may belong
            to the same experiment.
        :param experiment_index_list: Indexes indicating to which experiment
            the qubit pauli strings belong to.
        :return: A list of merged qubit pauli operators.
        """

        # A dictionary mapping the experiment index to the dictionary
        # of the merged qubit pauli operator.
        index_to_merged_qpo_dict: Dict[
            int, Dict[QubitPauliString, CoeffTypeAccepted]
        ] = {}

        # For each qubit pauli operator, sum the coefficients of the
        # qubit pauli strings it contains.
        for index, qpo in zip(experiment_index_list, qpo_list):
            index_qpo_dict = index_to_merged_qpo_dict.get(
                index, {qps: 0 for qps in qpo._dict.keys()}
            )

            if not index_qpo_dict.keys() == qpo._dict.keys():
                raise Exception(
                    "The qubit pauli strings being merged do not contain "
                    + "matching qubit pauli strings. In particular "
                    + f"{index_qpo_dict.keys()} and {qpo._dict.keys()} differ."
                )

            # Sum qubit pauli string coefficients.
            index_to_merged_qpo_dict[index] = {
                qps: qpo._dict[qps] + index_qpo_dict[qps] for qps in qpo._dict.keys()
            }

        # For each index, find the mean coefficient.
        for index, merged_qpo_dict in index_to_merged_qpo_dict.items():
            index_to_merged_qpo_dict[index] = {
                qps: coeff / experiment_index_list.count(index)
                for qps, coeff in merged_qpo_dict.items()
            }

        # Convert the dictionary to a list and return.
        return (
            [
                QubitPauliOperator(dictionary=index_to_merged_qpo_dict.get(index, {}))
                for index in range(max(experiment_index_list) + 1)
            ],
        )

    return MitTask(
        _label="MergeExperiments", _n_in_wires=2, _n_out_wires=1, _method=task
    )


def extrapolation_task_gen(
    noise_scaling_list: List[float], _fit_type: Fit, _show_fit: bool, deg: int
) -> MitTask:
    """Generates task extrapolating to the zero noise limit using results from many folded circuits.

    :param noise_scaling_list: A list of the values by which the noise has been folded.
    :param _fit_type: The function used to fit to the resulting data.
    :param _show_fit: Plot data and resulting fitted function.
    :param deg: The degree of polynomials used.
    """

    def task(
        obj, base_exp_list: List[QubitPauliOperator], *args
    ) -> Tuple[List[QubitPauliOperator]]:
        """Returns expectation values corrected by extrapolation

        :param base_exp_list: List of expectation values corresponding to each experiment
        :return: Each element of this tuple is a list of expectations for each
        experiment, all with noise scaled by a fixed amount.
        """

        # Reformats to create list, where each list has fixed noise folding. Each element of
        # the list is a list of experiments
        expanded_args = [i for i in args]

        # Reformats so that all_fold_qpo_list_floats is a list with each element
        # corresponding to a fixed experiment. Each element itself is a dictionary {qps:list}
        # where the list is a list of expectation values from each noise folded experiment
        all_fold_qpo = [qpo for qpo in zip(base_exp_list, *expanded_args)]
        all_fold_qpo_list_floats = [
            {k: [d[k] for d in qpo_tuple] for k in qpo_tuple[0]._dict.keys()}
            for qpo_tuple in all_fold_qpo
        ]

        extrapolated = []

        # The list of expectations for each experiment is now used to
        # extrapolate to the ideal value
        for qpo_list_float in all_fold_qpo_list_floats:
            extrapolated.append(
                QubitPauliOperator(
                    {
                        qpo_k: _fit_type(  # type: ignore
                            noise_scaling_list, qpo_list_float[qpo_k], _show_fit, deg
                        )
                        for qpo_k in qpo_list_float
                    }
                )
            )

        return (extrapolated,)

    return MitTask(
        _label="CollateZNEResults",
        _n_in_wires=len(noise_scaling_list),
        _n_out_wires=1,
        _method=task,
    )


def copy_mitex_wire(wire: ObservableExperiment) -> ObservableExperiment:
    """Returns a single copy of the inputted wire

    :param wire: Pair of ansatz circuit and ObservableTracker
    :return: single copy of inputted wire
    """

    # Copy ansatz circuit
    new_ansatz_circuit = AnsatzCircuit(
        Circuit=deepcopy(wire.AnsatzCircuit.Circuit),
        Shots=deepcopy(wire.AnsatzCircuit.Shots),
        SymbolsDict=deepcopy(wire.AnsatzCircuit.SymbolsDict),
    )

    # copy qps and instantiate new measurement setup
    new_ot = ObservableTracker(
        QubitPauliOperator(copy(wire.ObservableTracker._qubit_pauli_operator._dict))
    )
    new_wire = ObservableExperiment(
        AnsatzCircuit=new_ansatz_circuit, ObservableTracker=new_ot
    )
    return new_wire


def gen_duplication_task(duplicates: int, **kwargs) -> MitTask:
    """Duplicate the inputted experiment wire

    :param duplicates: The number of times to duplicate the input wire.
    """

    def task(
        obj,
        mitex_wire: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment], ...]:
        """Duplicate the inputted experiment wire

        :param mitex_wire: List of experiments
        :raises ValueError: Raised if the number of duplications is less than 1
        :return: Many copies of the inputted wire
        """

        # Raise error if the number of duplications requested is less than 1
        if duplicates <= 0:
            raise ValueError(
                "The number of duplications must be greater than or equal to 1."
            )

        # Compose coppies into wire format
        if duplicates == 1:
            me_copy = [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
            return (me_copy,)
        else:
            me_copies = [
                [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
            ]
            for _ in range(duplicates - 1):
                mc = [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
                me_copies.append(mc)
            return tuple(me_copies)

    return MitTask(
        _label=kwargs.get("_label", "Duplicate"),
        _n_out_wires=duplicates,
        _n_in_wires=1,
        _method=task,
    )


def qpo_node_relabel(
    qpo: QubitPauliOperator,
    node_map: Dict[Union[UnitID, Qubit, Node], Union[UnitID, Qubit, Node]],
):
    """Relabel the nodes of qpo according to node_map

    :param qpo: Original qubit pauli operator
    :param node_map: Map between nodes
    :return: Relabeled qubit pauli operator
    """

    orig_qpo_dict = qpo._dict.copy()
    new_qpo_dict: Dict[QubitPauliString, Union[int, float, complex, Expr]] = {}
    for orig_qps in orig_qpo_dict:
        orig_qps_dict = orig_qps.map
        new_qps_dict = {}
        for q in orig_qps_dict:
            new_qps_dict[node_map[q]] = orig_qps_dict[q]
        new_qps = QubitPauliString(cast(Dict[Qubit, Pauli], new_qps_dict))
        new_qpo_dict[new_qps] = orig_qpo_dict[orig_qps]

    return QubitPauliOperator(new_qpo_dict)


def gen_initial_compilation_task(
    backend: Backend, optimisation_level: int = 1
) -> MitTask:
    """Perform compilation to the backend. Note that this will relabel the
    nodes of the device, and so should be followed by gen_qubit_relabel_task
    in the task graph.

    :param backend: Backend to compile to
    :param optimisation_level: level of default compiler, defaults to 1
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment], List[Dict[UnitID, UnitID]]]:
        """Performs initial compilation before folding. This is to ensure minimal compilation
        after folding, as this could disrupt by how much the noise is increased.

        :param wire: List of experiments
        :return: List of experiments compiled to run on the inputted backend.
            Additionally a list of dictionaries describing how the nodes have
            been mapped by compilation.
        """

        mapped_wire = []
        node_map_list = []

        for obs_exp in wire:
            # Perform default compilation, tracking to which physical
            # qubits the initial qubits are mapped
            compiled_circ = obs_exp.AnsatzCircuit.Circuit.copy()

            cu = CompilationUnit(compiled_circ)

            if backend._uses_lightsabre:
                backend.default_compilation_pass(  # type: ignore
                    optimisation_level=optimisation_level,
                    allow_symbolic=True,
                ).apply(cu)
            else:
                backend.default_compilation_pass(
                    optimisation_level=optimisation_level
                ).apply(cu)
            node_map = cu.final_map
            node_map_list.append(node_map)

            # Alter the qubit pauli operator so that it maps to the new physical qubits.
            qpo = obs_exp[1]._qubit_pauli_operator
            new_qpo = qpo_node_relabel(qpo, node_map)

            # Construct new list of experiments, but with compiled circuits.
            mapped_wire.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=cu.circuit,
                        Shots=obs_exp.AnsatzCircuit.Shots,
                        SymbolsDict=obs_exp.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=ObservableTracker(new_qpo),
                )
            )

        return (
            mapped_wire,
            node_map_list,
        )

    return MitTask(
        _label="CompileToBackend",
        _n_out_wires=2,
        _n_in_wires=1,
        _method=task,
    )


def gen_qubit_relabel_task() -> MitTask:
    """Task reversing the relabelling of qubits performed during compilation.
    This should follow gen_initial_compilation_task

    :return: Task performing relabelling.
    """

    def task(
        obj,
        qpo_list: List[QubitPauliOperator],
        compilation_map_list: List[Dict[Node, Node]],
    ) -> Tuple[List[QubitPauliOperator]]:
        """Use node map returned by compilation unit to undo the relabelling
        performed by gen_initial_compilation_task

        :param qpo_list: List of QubitPauliOperator
        :param compilation_map_list: List of Dictionaries mapping nodes as
            returned by gen_initial_compilation_task task
        :return: List of QubitPauliOperator with relabeled nodes.
        """

        new_qpo_list = []

        for compilation_map, qpo in zip(compilation_map_list, qpo_list):
            node_map = {value: key for key, value in compilation_map.items()}
            new_qpo_list.append(
                qpo_node_relabel(
                    qpo,
                    cast(
                        Dict[Union[UnitID, Qubit, Node], Union[UnitID, Qubit, Node]],
                        node_map,
                    ),
                )
            )

        return (new_qpo_list,)

    return MitTask(
        _label="RelabelQubits",
        _n_out_wires=1,
        _n_in_wires=2,
        _method=task,
    )


def gen_noise_scaled_mitex(
    backend: Backend,
    noise_scaling: float,
    **kwargs,
) -> MitEx:
    """Generates MitEx with noise scaled by the Qermit Folding methods.

    :param backend: Backend on which circuits are run.
    :param noise_scaling: Factor by which noise is scaled.

    :return: MitEx with scaled noise.

    :key experiment_mitres: MitRes on which circuits are run, defaults to
        a MitRes wrapped around the given backend.
    :key experiment_mitex: MitEx on which the circuits are run, defaults
        to a MitEx wrapped around experiment_mitres
    :key allow_approx_fold: Allow approximate folding which may occur as a
        result of discreet folding due to adding gates.
    :key folding_type: The noise scaling method to use.
    """

    _experiment_mitres = deepcopy(
        kwargs.get(
            "experiment_mitres",
            MitRes(backend),
        )
    )

    _experiment_mitex = deepcopy(
        kwargs.get(
            "experiment_mitex",
            MitEx(backend, _label="ExperimentMitex", mitres=_experiment_mitres),
        )
    )

    _allow_approx_fold = kwargs.get("allow_approx_fold", True)
    _folding_type = kwargs.get("folding_type", Folding.circuit)

    _fold_mitex = deepcopy(_experiment_mitex)
    _fold_mitex._label = str(noise_scaling) + "FoldMitEx" + _fold_mitex._label

    digital_folding_task = digital_folding_task_gen(
        backend, noise_scaling, _folding_type, _allow_approx_fold, **kwargs
    )

    _fold_taskgraph = TaskGraph().from_TaskGraph(_fold_mitex)
    _fold_taskgraph.add_wire()
    _fold_taskgraph.prepend(digital_folding_task)
    _fold_taskgraph.append(merge_experiments_task_gen())

    return MitEx(backend).from_TaskGraph(_fold_taskgraph)


# TODO: Backend does not appear as input in documentation
def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -> MitEx:
    """Generates MitEx object which mitigates for noise using Zero Noise Extrapolation. This is the
    process by which noise is amplified incrementally, and the zero noise case arrived at by
    extrapolating backwards. For further explanantion see https://arxiv.org/abs/2005.10921.

    :param backend: Backend on which the circuits are to be run.
    :param noise_scaling_list: A list of the amounts by which the noise should be scaled.
    :return: MitEx object performing noise mitigation by ZNE.
    """
    _optimisation_level = kwargs.get("optimisation_level", 0)
    _show_fit = kwargs.get("show_fit", False)
    _fit_type = kwargs.get("fit_type", Fit.linear)
    _deg = kwargs.get("deg", len(noise_scaling_list) - 1)
    _seed = kwargs.get("seed", None)

    np.random.seed(seed=_seed)

    _zne_taskgraph = TaskGraph().from_TaskGraph(
        gen_noise_scaled_mitex(
            noise_scaling=noise_scaling_list[0],
            backend=backend,
            **kwargs,
        )
    )

    for fold in noise_scaling_list[1:]:
        _zne_taskgraph.parallel(
            gen_noise_scaled_mitex(
                noise_scaling=fold,
                backend=backend,
                **kwargs,
            )
        )

    _zne_taskgraph.prepend(gen_duplication_task(len(noise_scaling_list)))
    _zne_taskgraph.append(
        extrapolation_task_gen(noise_scaling_list, _fit_type, _show_fit, _deg)
    )

    _zne_taskgraph.add_wire()

    _zne_taskgraph.prepend(gen_initial_compilation_task(backend, _optimisation_level))
    _zne_taskgraph.append(gen_qubit_relabel_task())

    return MitEx(backend).from_TaskGraph(_zne_taskgraph)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import warnings
from abc import ABC, abstractmethod
from copy import copy
from typing import Dict, List, Tuple, Union, cast

import numpy as np
from pytket.pauli import QubitPauliString
from pytket.utils import QubitPauliOperator
from sympy.core.expr import Expr  # type: ignore

from qermit import MitTask


class _BaseExCorrectModel(ABC):
    """
    Model for storing calibrations and correcting expectation values for some experiments.
    """

    @abstractmethod
    def calibrate(self, noisy_exp: List[float], exact_exp: List[float]) -> None:
        pass

    @abstractmethod
    def correct(self, noisy_expectation: float) -> float:
        return NotImplemented


class _PolyCDRCorrect(_BaseExCorrectModel):
    def __init__(self, degree: int, params: List[float] = []) -> None:
        super().__init__()
        if params:
            if len(params) != degree + 1:
                raise ValueError(
                    f"Degree {degree} polynomial requires {degree + 1} parameters, "
                    f"{len(params)} were provided."
                )
        self.degree = degree
        self.params = params

    def calibrate(self, noisy_exp: List[float], exact_exp: List[float]) -> None:
        self.params = cast(
            List[float],
            cast(
                np.ndarray, np.polyfit(x=noisy_exp, y=exact_exp, deg=self.degree)
            ).tolist(),
        )

    def correct(self, noisy_expectation: float) -> float:
        return cast(
            float,
            sum(
                coef * (noisy_expectation**power)
                for coef, power in zip(self.params, range(self.degree, -1, -1))
            ),
        )


def cdr_quality_check_task_gen(
    distance_tolerance: float, calibration_fraction: float
) -> MitTask:
    """
    Check quality of calibration results. In particular, ensure that a
    significant proportion of the calibrations circuits have noisy expectation
    value close to that of the noisy expectation value of the original circuit.

    :param distance_tolerance: The absolute tolerance on the distance between
    expectation values of the calibration and original circuit.
    :param calibration_fraction: The upper bound on the fraction of calibration
        circuits which have noisy expectation values far from that of the
        original circuit.
    """

    def cdr_quality_check_task(
        obj,
        noisy_expectation: List[QubitPauliOperator],
        state_circuit_exp: List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]],
    ) -> Tuple[
        List[QubitPauliOperator],
        List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]],
    ]:
        """
        For each calibration result, check the difference between its noisy
        expectation value and that of the original circuit. Raise a warning if
        this value is large for a significant portion of the calibration
        circuits.

        :param noisy_expectation: List of noisy expectation values from
        original circuits.
        :param state_circuit_exp: A list of noisy and noiseless expectation
        values for each calibration experiment.
        :return: The original inputs are returned unaltered.
        """

        for calibration, original in zip(state_circuit_exp, noisy_expectation):
            # The noisy expectation value of the original circuit.
            original_coefficient = original.to_list()[0]["coefficient"][0]

            is_far_count = 0
            for qpo_pair in calibration:
                noisy_qpo = qpo_pair[0]
                # The noisy expectation value of the calibration circuit.
                noisy_coefficient = noisy_qpo.to_list()[0]["coefficient"][0]

                # Check the closeness of the expectation value of the training
                # circuit and that of the original circuit.
                if not math.isclose(
                    noisy_coefficient, original_coefficient, abs_tol=distance_tolerance
                ):
                    is_far_count += 1

            # Raise a warning if the calibration circuits regularly have noisy
            # expectation value far from the original circuit.
            if is_far_count > len(calibration) * calibration_fraction:
                warnings.warn(
                    "Training data regularly differs significantly from original circuit, fit and results may be poor."
                )

        return (
            noisy_expectation,
            state_circuit_exp,
        )

    return MitTask(
        _label="CDRQualityCheck",
        _n_in_wires=2,
        _n_out_wires=2,
        _method=cdr_quality_check_task,
    )


def cdr_calibration_task_gen(
    model: _BaseExCorrectModel,
) -> MitTask:
    """
    Uses calibration results from running characterisation circuits through a device
    and a noiseless simulator to characterise some model for correcting noisy expectation values.

    :param backend: Backend for storing characterisation model in.
    :param model: Model type to be calibrated and stored in backend.
    """

    def cdr_calibration_task(
        obj,
        calibration_results: List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]],
    ) -> Tuple[bool]:
        """
        For each experiments calibration results, calibrates a model of model type and stores it in backend.

        :param calibration_results: A list of noisy and noiseless expectation values for each
        experiment.

        :return: A bool confirming characterisation is complete
        """
        counter = 0
        for calibration in calibration_results:
            # dict from QubitPauliString to Tuple[List[float], List[float]]
            # facilitates characterisation of different QubitPauliStrings from different experiments
            noisy_char_dict: Dict[QubitPauliString, List[float]] = dict()
            exact_char_dict: Dict[QubitPauliString, List[float]] = dict()

            for qpo_pair in calibration:
                noisy_qpo = qpo_pair[0]
                exact_qpo = qpo_pair[1]

                # go through strings in operator
                for key in noisy_qpo._dict:
                    # make sure keys are present (don't initialise at start incase indexing missing)
                    if key not in noisy_char_dict:
                        noisy_char_dict[key] = list()
                    if key not in exact_char_dict:
                        exact_char_dict[key] = list()
                    if key not in exact_qpo._dict:
                        raise ValueError(
                            "Given key in calibration task for Clifford Data Regression should be present in exact and noisy characterisation results."
                        )

                    noisy_char_dict[key].append(float(noisy_qpo._dict[key]))
                    exact_char_dict[key].append(float(exact_qpo._dict[key]))
            obj.characterisation["CDR_" + str(counter)] = dict()
            # for each qubit pauli string in operator, add model for calibrating
            for key in noisy_char_dict:
                model.calibrate(noisy_char_dict[key], exact_char_dict[key])
                # calibrate creates new model, copy it in to backend
                obj.characterisation["CDR_" + str(counter)][key] = copy(model)
            counter += 1
        return (True,)

    return MitTask(
        _label="CDRCalibrate",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=cdr_calibration_task,
    )


def cdr_correction_task_gen() -> MitTask:
    """
    For each QubitPauliOperator passed, corrects the given expectation for each
    internal QubitPauliString via some pre-characterised mode.

    :param backend: Backend object where characterisation is stored.
    """

    def cdr_correction_task(
        obj,
        calibration_complete: bool,
        noisy_expectation: List[QubitPauliOperator],
    ) -> Tuple[List[QubitPauliOperator]]:
        """
        :param noisy_expectation: QubitPauliOperator objects from some experiment
        which are to be corrected from some predefined characteriastion.
        :param calibration_complete: bool passed to method once calibration task has completed.

        """
        corrected_expectations = []
        for i in range(len(noisy_expectation)):
            models = obj.characterisation["CDR_" + str(i)]
            new_qpo_dict = dict()
            for qps in noisy_expectation[i]._dict:
                if qps in models:
                    new_qpo_dict[qps] = cast(
                        Union[int, float, complex, Expr],
                        models[qps].correct(float(noisy_expectation[i]._dict[qps])),
                    )
                else:
                    new_qpo_dict[qps] = cast(
                        Union[int, float, complex, Expr],
                        noisy_expectation[i]._dict[qps],
                    )
            corrected_expectations.append(QubitPauliOperator(new_qpo_dict))

        return (corrected_expectations,)

    return MitTask(
        _label="CDRCorrect",
        _n_in_wires=2,
        _n_out_wires=1,
        _method=cdr_correction_task,
    )



from copy import deepcopy
from typing import List, Tuple, cast

from pytket.backends import Backend
from pytket.backends.backendinfo import BackendInfo
from pytket.extensions.quantinuum.backends.leakage_gadget import get_detection_circuit

from qermit import CircuitShots, MitRes, MitTask, TaskGraph
from qermit.postselection import PostselectMgr
from qermit.postselection.postselect_mitres import gen_postselect_task


def gen_add_leakage_gadget_circuit_task(backend: Backend) -> MitTask:
    """Generates task adding leakage gadget circuits to given circuts.

    :param backend: Backend on which the circuit will be run.
    :return: Task adding leakage gadget circuits to given circuts.
    """

    if backend.backend_info is None:
        raise Exception("This backend has no nodes.")

    n_device_qubits = cast(BackendInfo, backend.backend_info).n_nodes

    def task(
        obj, circuit_shots_list: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[PostselectMgr]]:
        """Task adding leakage gadget circuits to given circuts. This reuses
        methods from pytket-quantinuum. A list of the corresponding
        postselection managers is also created.

        :param circuit_shots_list: List of circuits to which leakage gadget
            circuit should be added.
        :return: Circuits with gadget added, and list of corresponding
            post selection managers.
        """

        # Add leakage detection gadget to each inputted circuit.
        detection_circuit_shots_list = [
            CircuitShots(
                Circuit=get_detection_circuit(
                    circuit=circuit_shots.Circuit,
                    n_device_qubits=n_device_qubits,
                ),
                Shots=circuit_shots.Shots,
            )
            for circuit_shots in circuit_shots_list
        ]

        # For each circuit create a postselection manager. These may be
        # different for each circuit, if for example the circuits are of
        # different sizes.
        postselect_mgr_list = [
            PostselectMgr(
                compute_cbits=orig_circuit.Circuit.bits,
                postselect_cbits=list(
                    set(detection_circuit.Circuit.bits).difference(
                        set(orig_circuit.Circuit.bits)
                    )
                ),
            )
            for orig_circuit, detection_circuit in zip(
                circuit_shots_list, detection_circuit_shots_list
            )
        ]

        return (
            detection_circuit_shots_list,
            postselect_mgr_list,
        )

    return MitTask(
        _label="AddLeakageGadget",
        _n_in_wires=1,
        _n_out_wires=2,
        _method=task,
    )


def get_leakage_detection_mitres(backend: Backend, **kwargs) -> MitRes:
    """Generate MitRes making use of leakage detection and postselection.

    :param backend: Backend on which the circuits are run.
    :return: MitRes making use of leakage detection and postselection.
    """

    _mitres = deepcopy(
        kwargs.get("mitres", MitRes(backend, _label="LeakageDetectionMitRes"))
    )
    _taskgraph = TaskGraph().from_TaskGraph(_mitres)
    _taskgraph.add_wire()
    # Prepend task adding leakage detection circuits.
    _taskgraph.prepend(gen_add_leakage_gadget_circuit_task(backend))
    # Append task removing shots where leakage is detected
    _taskgraph.append(gen_postselect_task())

    return MitRes(backend).from_TaskGraph(_taskgraph)



from copy import deepcopy
from typing import List, Tuple

from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult

from qermit import CircuitShots, MitRes, MitTask, TaskGraph

from .postselect_manager import PostselectMgr


def gen_postselect_task() -> MitTask:
    """Generates task applying postselection to given results.

    :return: Task applying postselection to given results.
    """

    def task(
        obj,
        result_list: List[BackendResult],
        postselect_mgr_list: List[PostselectMgr],
    ) -> Tuple[List[BackendResult]]:
        """Task applying postselection to given results.

        :param result_list: List od results to which postselection should
            be applied.
        :param postselect_mgr_list: List of postselection managers to apply
            to results.
        :return: List of results after postselection has been applied.
        """

        return (
            [
                postselect_mgr.postselect_result(result)
                for result, postselect_mgr in zip(result_list, postselect_mgr_list)
            ],
        )

    return MitTask(
        _label="PostselectResults",
        _n_in_wires=2,
        _n_out_wires=1,
        _method=task,
    )


def gen_postselect_mgr_gen_task(postselect_mgr: PostselectMgr) -> MitTask:
    """Generates task applying the same post selection manager to all
    circuits.

    :param postselect_mgr: Postselection manager to apply.
    :return: Task applying the same post selection manager to all circuits.
    """

    def task(
        obj, circ_shots_list: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[PostselectMgr]]:
        """Task applying the same post selection manager to all circuits.

        :param circ_shots_list: List of circuits to which post selection is
            applied.
        :return: List od circuits and corresponding postselection managers.
        """

        return (circ_shots_list, [postselect_mgr for _ in circ_shots_list])

    return MitTask(
        _label="UniformPostselection",
        _n_in_wires=1,
        _n_out_wires=2,
        _method=task,
    )


def gen_postselect_mitres(
    backend: Backend, postselect_mgr: PostselectMgr, **kwargs
) -> MitRes:
    """Generates MitRes running given circuit and applying postselection.

    In the following example we prepare and measure a Bell state.

    .. jupyter-execute::

        from pytket import Circuit
        from pytket.circuit.display import render_circuit_jupyter

        circuit = Circuit(2,2).H(0).CX(0,1).measure_all()
        render_circuit_jupyter(circuit)

    We would like to postselect one measurement outcome based on the
    other being 0. We prepare a postselect mitres accordingly.

    .. jupyter-execute::

        from qermit.postselection import gen_postselect_mitres
        from pytket.extensions.quantinuum import QuantinuumBackend, QuantinuumAPIOffline
        from qermit.postselection import PostselectMgr

        backend = QuantinuumBackend(
            device_name="H1-1LE",
            api_handler = QuantinuumAPIOffline()
        )

        postselect_mgr = PostselectMgr(
            compute_cbits=[circuit.bits[0]],
            postselect_cbits=[circuit.bits[1]]
        )

        mitres = gen_postselect_mitres(
            backend=backend,
            postselect_mgr=postselect_mgr,
        )
        mitres.get_task_graph()

    We can then construct the experiment we wish to run, and pass it through
    the postselect mitres.

    .. jupyter-execute::

        from qermit import CircuitShots

        circ_shots = CircuitShots(
            Circuit=backend.get_compiled_circuit(circuit),
            Shots=100
        )
        result_list = mitres.run([circ_shots])
        result_list[0].get_counts()

    :param backend: Backend on this circuits are run.
    :param postselect_mgr: Postselection manager.
    :return: MitRes running given circuit and applying postselection.
    """

    _mitres = deepcopy(
        kwargs.get("mitres", MitRes(backend, _label="PostselectionMitRes"))
    )
    _taskgraph = TaskGraph().from_TaskGraph(_mitres)
    _taskgraph.add_wire()
    _taskgraph.prepend(gen_postselect_mgr_gen_task(postselect_mgr))
    _taskgraph.append(gen_postselect_task())

    return MitRes(backend).from_TaskGraph(_taskgraph)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import copy
from typing import Dict, List, Tuple

from pytket import Bit
from pytket.backends import Backend
from pytket.partition import (
    GraphColourMethod,
    PauliPartitionStrat,
    measurement_reduction,
)
from pytket.pauli import QubitPauliString
from pytket.transform import CXConfigType

from .mitex import MitEx
from .mittask import MitTask, ObservableExperiment
from .utils import MeasurementCircuit


def measurement_reduction_task_gen(
    strategy: PauliPartitionStrat, method: GraphColourMethod, cx_config: CXConfigType
) -> MitTask:
    """
    Uses measurement reduction techniques available in the pytket.partition module to reduce
    the number of measurement circuits required to measure all expectations for an observable.

    :param strategy: Measurement reduction strategy used in measurement_reduction method
    :param method: Graph colouring method used in measurement_reduction method
    :param cx_config: Configuration of CX gates for diagonlisation methods.

    :return: MitTask object, taking and return same write type, but with added
        measurement circuits to ObservableTracker.
    """

    def task(
        obj,
        measurement_wires: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment]]:
        new_wires = []
        # for each experiment
        for measurement_wire in measurement_wires:
            # circ is ansatz circuit which measurement circuits are appended to
            circ = measurement_wire[0][0]
            # symbols dict object
            symbols = measurement_wire[0][2]
            # ObservableTracker object for given experiment
            tracker = measurement_wire[1]
            # use measurement reduction method to get MeasurementSetup object
            # with reduced number of measurement circuits
            new_setup = measurement_reduction(
                list(tracker._qubit_pauli_operator._dict.keys()),
                strat=strategy,
                method=method,
                cx_config=cx_config,
            )
            # attach measurement configurations from MeasurementSetup to
            # copies of AnsatzCircuit
            # new measurement circuits
            measurement_circuits = []
            for measurement_circuit in new_setup.measurement_circs:
                full_circ = circ.copy()
                full_circ.append(measurement_circuit)
                measurement_circuits.append(MeasurementCircuit(full_circ, symbols))
            # convert MeasurementBitMap objects to MeasurementInfo for ObservableTracker
            adder_info: Dict[int, List[Tuple[QubitPauliString, List[Bit], bool]]] = (
                dict()
            )
            for i in range(len(measurement_circuits)):
                adder_info[i] = list()

            for qps, value in new_setup.results.items():
                for v in value:
                    circ_index = v.circ_index
                    bits = [Bit(i) for i in v.bits]
                    invert = v.invert
                    adder_info[circ_index].append((qps, bits, invert))

            for key in adder_info:
                tracker.add_measurement_circuit(
                    measurement_circuits[key], adder_info[key]
                )

            new_wires.append(
                ObservableExperiment(
                    AnsatzCircuit=measurement_wire[0], ObservableTracker=tracker
                )
            )
        return (new_wires,)

    return MitTask(
        _label="MeasurementReduction", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def gen_MeasurementReduction_MitEx(backend: Backend, **kwargs) -> MitEx:
    """
    Returns a MitEx object that produces measurement circuits via Measurement Reduction
    methods.

    :param backend: Backend experiment is built around.
    :key mitex: MitEx object measurement reduction task is prepended to.

    :return: MitEx object performing Measurement reduction
    """
    # mitex object to built measurement reduction
    _mitex = copy(
        kwargs.get("mitex", MitEx(backend, _label="MeasurementReductionMitEx"))
    )
    _mitex._label = "MeasurementReductionMitEx"
    _strategy = kwargs.get("strategy", PauliPartitionStrat.NonConflictingSets)
    _method = kwargs.get("method", GraphColourMethod.Lazy)
    _cx_config = kwargs.get("cx_config", CXConfigType.Snake)
    # add measurement reduction task to start of method
    _mitex.prepend(measurement_reduction_task_gen(_strategy, _method, _cx_config))
    return _mitex



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from enum import Enum
from types import MethodType
from typing import Callable, Dict, List, Optional, Union

from pytket import Bit, Circuit, Qubit
from pytket.backends import ResultHandle
from pytket.backends.backendresult import BackendResult
from pytket.utils import QubitPauliOperator


class IOTask(Enum):
    """
    Simple Node type for labelling the Input and Output Nodes to a TaskGraph object.
    """

    Input = 0
    Output = 1


CircuitShots = namedtuple("CircuitShots", ["Circuit", "Shots"])
AnsatzCircuit = namedtuple("AnsatzCircuit", ["Circuit", "Shots", "SymbolsDict"])
ObservableExperiment = namedtuple(
    "ObservableExperiment", ["AnsatzCircuit", "ObservableTracker"]
)

Wire = Union[
    CircuitShots,
    Circuit,
    BackendResult,
    ResultHandle,
    AnsatzCircuit,
    ObservableExperiment,
    int,
    float,
    bool,
    str,
    QubitPauliOperator,
    Dict[Qubit, Bit],
    Dict,
]


class MitTask:
    """
    An object a TaskGraph node is comprised of. A MitTask object
    is defined by the _method attribute, which holds a pure function
    that requires _n_in_wires input arguments and returns a Tuple of
    _n_out_wires objects. The object callable is defined as the
    _method attribute.

    :param _label: String to identify MitTask object by.
    :param _n_in_wires: Number of input arguments to _method attribute function.
    :param _n_out_wires: number of results in Tuple returned by _method attribute function.
    :param _method: Pure function executed when object called.

    :return: MitTask object for adding to TaskGraph.
    """

    def __init__(
        self,
        _label: str,
        _n_in_wires: int,
        _n_out_wires: int,
        _method: Optional[Callable] = None,
    ):
        self._label = _label
        self._n_in_wires = _n_in_wires
        self._n_out_wires = _n_out_wires
        if _method:
            self.run = MethodType(_method, self)
        self.characterisation: dict = {}

    @property
    def label(self) -> str:
        return self._label

    @property
    def n_in_wires(self):
        return self._n_in_wires

    @property
    def n_out_wires(self):
        return self._n_out_wires

    def __call__(self, input_wires: List[Wire]) -> List[Wire]:
        return self.run(*input_wires)

    def __str__(self):
        return f"<MitTask::{self._label}>"

    def __repr__(self):
        return str(self)


def duplicate_wire_task_gen(in_wires: int, duplicates: int) -> MitTask:
    """
    Generator for constructing a task that for each argument, corresponding
    to data from some input edge on graph, makes duplicates copies to be returned in Tuple
    (and later added to out edges of MitTask on TaskGraph).

    :param in_wires: Number of in edges to Task on TaskGraph.
    :param duplicates: Number of copies to take of each argument.
    """

    def task(obj, *args):
        if len(args) != in_wires:
            raise ValueError(
                "Task has "
                + str(len(args))
                + " input arguments but expects "
                + str(in_wires)
                + "."
            )
        return args * duplicates

    return MitTask(
        _label="Duplicate" + str(in_wires) + "Wires" + str(duplicates) + "Times",
        _n_in_wires=in_wires,
        _n_out_wires=in_wires * duplicates,
        _method=task,
    )



from __future__ import annotations

from typing import List

from pytket.circuit import Circuit, Op, OpType, Qubit
from pytket.pauli import Pauli, QubitPauliString, QubitPauliTensor, pauli_string_mult
from pytket.tableau import UnitaryTableau


class QermitPauli:
    """For the manipulation of Pauli strings. In particular, how they are
    changed by the action of Clifford circuits.
    """

    coeff_to_phase = {1 + 0j: 0, 0 + 1j: 1, -1 + 0j: 2, 0 - 1j: 3}

    def __init__(self, qpt: QubitPauliTensor) -> None:
        """Initialised with an initial qubit pauli tensor. Other methods will
        modify this Qubit Pauli Tensor.

        :param qpt: Initial Qubit Pauli Tensor.
        """
        self.qubit_list = list(qpt.string.map.keys())

        self.unitary_tableau = UnitaryTableau(nqb=len(self.qubit_list))

        self.qubit_index = {qubit: Qubit(i) for i, qubit in enumerate(self.qubit_list)}
        self.index_qubit = {Qubit(i): qubit for i, qubit in enumerate(self.qubit_list)}

        self.input_pauli_tensor = QubitPauliTensor(
            string=QubitPauliString(
                map={
                    self.qubit_index[qubit]: pauli
                    for qubit, pauli in qpt.string.map.items()
                }
            ),
            coeff=qpt.coeff,
        )

    def is_measureable(self, qubit_list: List[Qubit]) -> bool:
        """Checks if this Pauli would be measurable on the given qubits in the
        computational bases. That is to say if at least one  Pauli on the given
        qubits anticommutes with Z.

        :param qubit_list: Qubits on which if measurable should be checked.
        :raises Exception: Raised if the given qubits are not contained
            in this Pauli.
        :return: True if at least one Pauli on the given
            qubits anticommutes with Z. False otherwise.
        """
        return any(
            not self.qubit_pauli_tensor.commutes_with(
                QubitPauliTensor(qubit=qubit, pauli=Pauli.Z)
            )
            for qubit in qubit_list
        )

    def reduce_qubits(self, qubit_list: List[Qubit]) -> QermitPauli:
        """Reduces Pauli by removing terms acting on qubits
        in the given list. A new reduced Pauli is created.

        :param qubit_list: Qubits in Pauli which should be removed.
        :return: Reduced Pauli.
        """
        return QermitPauli(
            QubitPauliTensor(
                string=QubitPauliString(
                    map={
                        qubit: pauli
                        for qubit, pauli in self.qubit_pauli_tensor.string.map.items()
                        if qubit not in qubit_list
                    }
                ),
                coeff=self.qubit_pauli_tensor.coeff,
            )
        )

    @property
    def dagger(self) -> QermitPauli:
        """Generates the inverse of the Pauli.

        :return: Conjugate transpose of the Pauli.
        """
        return QermitPauli(
            qpt=QubitPauliTensor(
                string=self.qubit_pauli_tensor.string,
                coeff=self.qubit_pauli_tensor.coeff.conjugate(),
            )
        )

    def __hash__(self) -> int:
        return self.qubit_pauli_tensor.__hash__()

    def __str__(self) -> str:  # pragma: no cover
        return str(self.qubit_pauli_tensor.string) + str(self.qubit_pauli_tensor.coeff)

    def __eq__(self, other: object) -> bool:
        """Checks for equality by checking all qubits match, and that all
        Paulis on those qubits match.

        :param other: Pauli to compare against.
        :return: True is equivalent.
        """

        if not isinstance(other, QermitPauli):
            return False

        return self.qubit_pauli_tensor == other.qubit_pauli_tensor

    def apply_circuit(self, circuit: Circuit) -> None:
        """Commute the Pauli through the circuit.
        Given a Clifford circuit C, transform the Pauli P to Q
        such that PC = CQ.

        :param circuit: Circuit to be applied.
        """

        for command in circuit.get_commands():
            if command.op.type == OpType.Barrier:
                continue

            if command.qubits != command.args:
                raise Exception(
                    "Circuit must be purely quantum."
                    f"The given circuit acts on bits {command.bits}"
                )

            self.apply_gate(
                op=command.op,
                qubits=command.qubits,
            )

    def apply_gate(self, op: Op, qubits: List[Qubit]) -> None:
        """Commute the Pauli through the gate.
        Given a Clifford gate G, transform the Pauli P to Q
        such that PG = GQ.

        :param op: Operation to commute through.
        :param qubits: Qubit on which the operation acts.
        :raises Exception: Raised if the given gate is not Clifford.
        """
        if not op.is_clifford():
            raise Exception(f"{op} is not a Clifford operation.")

        if op.is_clifford_type():
            self.unitary_tableau.apply_gate_at_end(
                type=op.type,
                qbs=[self.qubit_index[qubit] for qubit in qubits],
            )

        elif op.type == OpType.Rz:
            for _ in range(int((op.params[0] % 2) // 0.5)):
                self.apply_gate(op=Op.create(OpType.S), qubits=qubits)

        elif op.type == OpType.Rx:
            self.apply_gate(op=Op.create(OpType.H), qubits=qubits)
            for _ in range(int((op.params[0] % 2) // 0.5)):
                self.apply_gate(op=Op.create(OpType.S), qubits=qubits)
            self.apply_gate(op=Op.create(OpType.H), qubits=qubits)

        elif op.type == OpType.PhasedX:
            self.apply_gate(op=Op.create(OpType.Rz, [-op.params[1]]), qubits=qubits)
            self.apply_gate(op=Op.create(OpType.Rx, [op.params[0]]), qubits=qubits)
            self.apply_gate(op=Op.create(OpType.Rz, [op.params[1]]), qubits=qubits)

        elif op.type == OpType.ZZMax:
            self.apply_gate(op=Op.create(OpType.CX), qubits=qubits)
            self.apply_gate(op=Op.create(OpType.S), qubits=[qubits[1]])
            self.apply_gate(op=Op.create(OpType.CX), qubits=qubits)

        elif op.type == OpType.ZZPhase:
            for _ in range(int((op.params[0] % 2) // 0.5)):
                self.apply_gate(op=Op.create(OpType.ZZMax), qubits=qubits)

        else:
            raise NotImplementedError(
                f"{op} if clifford but is not supported. "
                "Please request the developers support this operation."
            )

    def pre_apply_pauli(self, pauli: Pauli, qubit: Qubit) -> None:
        """Transform Pauli P into PQ, where Q is the give pauli.

        :param pauli: Pauli to multiply by.
        :param qubit: Qubit on which pauli to multiply by should act.
        """
        mult_string, mult_coeff = pauli_string_mult(
            qubitpaulistring1=self.qubit_pauli_tensor.string,
            qubitpaulistring2=QubitPauliString(qubit=qubit, pauli=pauli),
        )
        mult_string_map = {
            self.qubit_index[qubit]: pauli for qubit, pauli in mult_string.map.items()
        }
        mult_string = QubitPauliString(map=mult_string_map)

        self.input_pauli_tensor = QubitPauliTensor(
            string=mult_string, coeff=self.qubit_pauli_tensor.coeff * mult_coeff
        )

        self.unitary_tableau = UnitaryTableau(nqb=len(self.qubit_list))

    def post_apply_pauli(self, pauli: Pauli, qubit: Qubit) -> None:
        """Transform Pauli P into QP, where Q is the give pauli.

        :param pauli: Pauli to multiply by.
        :param qubit: Qubit on which pauli to multiply by should act.
        """
        mult_string, mult_coeff = pauli_string_mult(
            qubitpaulistring1=QubitPauliString(qubit=qubit, pauli=pauli),
            qubitpaulistring2=self.qubit_pauli_tensor.string,
        )
        mult_string_map = {
            self.qubit_index[qubit]: pauli for qubit, pauli in mult_string.map.items()
        }
        mult_string = QubitPauliString(map=mult_string_map)

        self.input_pauli_tensor = QubitPauliTensor(
            string=mult_string, coeff=self.qubit_pauli_tensor.coeff * mult_coeff
        )

        self.unitary_tableau = UnitaryTableau(nqb=len(self.qubit_list))

    def get_control_circuit(self, control_qubit: Qubit) -> Circuit:
        """Controlled circuit which acts Pauli.

        :param control_qubit: Qubit on which circuit is controlled.
        :return: Controlled circuit acting Pauli.
        """
        circ = Circuit()
        circ.add_qubit(control_qubit)
        # TODO: in the case that this is secretly a controlled Y a controlled
        # Y should be applied. Otherwise there is additional noise added in
        # the case of a CY.
        phase = self.coeff_to_phase[self.qubit_pauli_tensor.coeff]
        for qubit, pauli in self.qubit_pauli_tensor.string.map.items():
            circ.add_qubit(id=qubit)

            if pauli == Pauli.Z or pauli == Pauli.Y:
                circ.CZ(
                    control_qubit=control_qubit,
                    target_qubit=qubit,
                    opgroup="pauli check",
                )

            if pauli == Pauli.X or pauli == Pauli.Y:
                circ.CX(
                    control_qubit=control_qubit,
                    target_qubit=qubit,
                    opgroup="pauli check",
                )

            if pauli == Pauli.Y:
                phase += 1
                phase %= 4

        for _ in range(phase):
            circ.S(
                control_qubit,
                opgroup="phase correction",
            )

        return circ

    @property
    def circuit(self) -> Circuit:
        """Circuit which acts Pauli.

        :return: Circuit acting Pauli.
        """
        circ = Circuit()

        phase = self.coeff_to_phase[self.qubit_pauli_tensor.coeff]

        for qubit, pauli in self.qubit_pauli_tensor.string.map.items():
            circ.add_qubit(id=qubit)

            if pauli == Pauli.Z or pauli == Pauli.Y:
                circ.Z(qubit)

            if pauli == Pauli.X or pauli == Pauli.Y:
                circ.X(qubit)

            if pauli == Pauli.Y:
                phase += 1
                phase %= 4

        circ.add_phase(a=phase / 2)

        return circ

    @property
    def qubit_pauli_tensor(self) -> QubitPauliTensor:
        """Current state of the Pauli, following any operations
        which may have been acted upon it until now.

        :return: Current state of the Pauli.
        """
        mislabled_qpt = self.unitary_tableau.get_row_product(
            paulis=self.input_pauli_tensor
        )
        correct_map = {
            qubit: mislabled_qpt.string.map.get(index, Pauli.I)
            for index, qubit in self.index_qubit.items()
        }
        return QubitPauliTensor(
            string=QubitPauliString(
                map=correct_map,
            ),
            coeff=mislabled_qpt.coeff,
        )



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
from copy import deepcopy
from typing import List, OrderedDict, Tuple, Union, cast

import networkx as nx  # type: ignore
from pytket import Bit, Circuit
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.pauli import Pauli, QubitPauliString
from pytket.utils import QubitPauliOperator

from .mitres import (
    MitRes,
    backend_compile_circuit_shots_task_gen,
    gen_shot_split_MitRes,
)
from .mittask import (
    AnsatzCircuit,
    CircuitShots,
    IOTask,
    MitTask,
    ObservableExperiment,
    Wire,
)
from .task_graph import TaskGraph
from .utils import MeasurementCircuit, MeasurementInfo, ObservableTracker, SymbolsDict


def gen_compiled_shot_split_MitRes(
    backend: Backend, max_shots: int, optimisation_level: int = 1
) -> MitRes:
    """
    Returns a shot splitting MitRes object with a compilation task prepended that compiles
    circuit wires via backend.compile_circuit. Optimisation level can be optionally
    set as defined by backend.compile_circuit.

    :param backend: Backend with circuits are compiled for.
    :param max_shots: The maximum number of shots that each job should request.
    :param optimisation_level: Sets options in compile_circuit method

    :return: shot splitting MitRes object with compilation task prepended.
    """
    mr = gen_shot_split_MitRes(backend, max_shots)
    mr.prepend(backend_compile_circuit_shots_task_gen(backend, optimisation_level))
    return mr


def gen_compiled_MitRes(backend: Backend, optimisation_level: int = 1) -> MitRes:
    """
    Returns a MitRes object with a compilation task prepended that compiles
    circuit wires via backend.compile_circuit. Optimisaion level can be optionally
    set as defined by backend.compile_circuit.

    :param backend: Backend with circuits are compiled for.
    :param optimisation_level: Sets options in compile_circuit method

    :return: MitRes object with compilation task prepended.
    """
    mr = MitRes(backend)
    mr.prepend(backend_compile_circuit_shots_task_gen(backend, optimisation_level))
    return mr


def get_basic_measurement_circuit(
    string: QubitPauliString,
) -> Tuple[Circuit, MeasurementInfo]:
    """
    Given a Qubit Pauli String, returns a circuit for measuring qubits in given basis
    via changing of basis through quantum gates.

    :param string: Qubit Pauli String to be measured

    :return: Measurement circuit for appending on some ansatz
    """
    measurement_circuit = Circuit()
    measured_qbs = []
    for qb, p in string.map.items():
        if p == Pauli.I:
            continue
        measured_qbs.append(qb)
        measurement_circuit.add_qubit(qb)
        if p == Pauli.X:
            measurement_circuit.H(qb)
        elif p == Pauli.Y:
            measurement_circuit.Rx(0.5, qb)
    bits = []
    for b_idx, qb in enumerate(measured_qbs):
        unit = Bit(b_idx)
        bits.append(unit)
        measurement_circuit.add_bit(unit, False)
        measurement_circuit.Measure(qb, unit)
    return (measurement_circuit, (string, bits, False))


def filter_observable_tracker_task_gen() -> MitTask:
    """
    Generates basic (changing measurement basis via Pauli gates) MeasurementCircuit
    for every QubitPauliString passed that has no Measurementcircuit in ObservableTracker object passed on wire.

    :return: Pure function that adds MeasurementCircuit objects to ObservableTracker.
    """

    def task(
        obj,
        measurement_wires: List[ObservableExperiment],
    ) -> Tuple[List[List[CircuitShots]], List[ObservableTracker]]:
        """
        :param measurement_wires: Wires containing Circuit information and Observable information
        :returns: A modified ObservableTracker object, and a List of CircuitShots for each Observable measured
        """

        output_circuits = []
        output_trackers = []
        for measurement_wire in list(measurement_wires):
            # set variables
            ansatz_circuit = measurement_wire[0]
            circuit = ansatz_circuit[0]
            shots = ansatz_circuit[1]
            symbols = ansatz_circuit[2]
            observable_tracker = measurement_wire[1]

            non_existant_qubits = (
                observable_tracker.qubit_pauli_operator.all_qubits.difference(
                    set(circuit.qubits)
                )
            )
            # TODO: In a future refactor, this check should be done
            # inside ObservableExperiment.
            if len(non_existant_qubits) > 0:
                raise Exception(
                    f"ObservableTracker qubits {non_existant_qubits} are not found in circuit."
                )

            # first make sure all observable has some measurement circuit
            strings_for_circuits = observable_tracker.get_empty_strings()
            for string in strings_for_circuits:
                circ = circuit.copy()
                # tuple, first entry is measurement circuit for appending
                # second entry is MeasurementInfo for deriving expectation
                measurement_circuit = get_basic_measurement_circuit(string)
                circ.append(measurement_circuit[0])
                # add new circuit to observable tracker
                observable_tracker.add_measurement_circuit(
                    MeasurementCircuit(circ, symbols), [measurement_circuit[1]]
                )

            # retrieve all measurement circuits, substitute symbols
            output_circuits.append(
                [
                    CircuitShots(Circuit=mc.get_parametric_circuit(), Shots=shots)
                    for mc in observable_tracker.measurement_circuits
                ]
            )
            output_trackers.append(observable_tracker)
        return (output_circuits, output_trackers)

    return MitTask(
        _label="FilterObservableTracker", _n_in_wires=1, _n_out_wires=2, _method=task
    )


def collate_circuit_shots_task_gen() -> MitTask:
    """
    Each wire contains a single experiment with its own List of Circuits to run.
    To improve parallelisation (i.e. reduce queueing time), these lists are collated
    and queued for a device at the same time.
    """

    def task(
        obj, circuit_wires: List[List[CircuitShots]]
    ) -> Tuple[List[CircuitShots], List[int]]:
        """
        :param circuit_wires: Different lists of Circuit + Shots for different experiments

        :return: Different experiment circuits collated on one wire, and the sequential number of
        results for each experiment on a second wire.
        """
        collated_circuitshots = []
        lengths = []
        for wire in circuit_wires:
            lengths.append(len(wire))
            collated_circuitshots.extend(wire)
        return (collated_circuitshots, lengths)

    return MitTask(
        _label="CollateExperimentCircuits", _n_in_wires=1, _n_out_wires=2, _method=task
    )


def split_results_task_gen() -> MitTask:
    """
    Returned list of Results from MitRes object are for multiple experiments.
    This method generates a task that converts the list of results into a list
    of list of results, wherein each list is for a different MitEx experiment.
    """

    def task(
        obj, results: List[BackendResult], experiment_sizes: List[int]
    ) -> Tuple[List[List[BackendResult]]]:
        """
        :param results: All results returned from MitRes object
        :param experiment_sizes: The ordered number of results required for each experiment

        :return: All results split up into sublists for each MitEx experiment
        """
        lower_bound = 0
        split_results = []
        for size in experiment_sizes:
            upper_bound = lower_bound + size
            split_results.append(results[lower_bound:upper_bound])
            lower_bound = upper_bound
        return (split_results,)

    return MitTask(_label="SplitResults", _n_in_wires=2, _n_out_wires=1, _method=task)


def get_expectations_task_gen() -> MitTask:
    """
    Passes each set of experiment results to corresponding ObservableTracker method, returning
    a QubitPauliOperator object containing expectation values from Results multiplied by coefficients.
    """

    def task(
        obj, all_results: List[List[BackendResult]], trackers: List[ObservableTracker]
    ) -> Tuple[List[QubitPauliOperator]]:
        """
        :param all_results: All Results from MitRes split into sublists, one for each Observable experiment.
        :param trackers: All ObservableTrackers defining experiments passed to MitEx.

        :return: Each experiments expectation results in a QubitPauliOperator
        """
        if len(all_results) != len(trackers):
            raise ValueError(
                "{} results and {} observable trackers. Values should be identical.".format(
                    len(all_results), len(trackers)
                )
            )
        output_qpos = [
            observable.get_expectations(results)
            for observable, results in zip(trackers, all_results)
        ]
        for observable in trackers:
            observable.clear()
        return (output_qpos,)

    return MitTask(
        _label="GenerateExpectations", _n_in_wires=2, _n_out_wires=1, _method=task
    )


class MitEx(TaskGraph):
    """
    A TaskGraph extension for mitigation of expectation values for individual QubitPauliStrings
            contained in some ObservableTracker/QubitPauliOperator.
    """

    def __init__(self, backend: Backend, _label: str = "MitEx", **kwargs) -> None:
        """
        MitEx objects are defined by the backend object experiments are run through.
        However, as experiments run through some MitRes object, kwargs
        can be used to run through any mitres of choice.

        :param backend: Pytket backend default constructor which tasks are generated from.
        :param label: Name for identification of MitEx object.
        :key mitres: MitEx object experiments are run through
        """
        # set member variables
        self._label = _label
        self._n_wires = 1
        self.G = None
        self.characterisation: dict = {}

        # start building default MitEx task graph
        self._task_graph = nx.MultiDiGraph()
        self._i, self._o = IOTask.Input, IOTask.Output

        # if requested, all data is held in cache and can be accessed after running
        self._cache: OrderedDict[str, Tuple[MitTask, List[Wire]]] = OrderedDict()

        # add edge from input to filtering task to generate measurement circuits
        filter_observable_tracker_task = filter_observable_tracker_task_gen()
        self._task_graph.add_edge(
            self._i, filter_observable_tracker_task, key=(0, 0), data=None
        )

        collate_circuit_shots_task = collate_circuit_shots_task_gen()
        self._task_graph.add_edge(
            filter_observable_tracker_task,
            collate_circuit_shots_task,
            key=(0, 0),
            data=None,
        )

        # if mitres isn't defined, build around a mitres which compiles circuits
        _mitres = kwargs.get("mitres", gen_compiled_MitRes(backend))
        self._task_graph.add_edge(
            collate_circuit_shots_task, _mitres, key=(0, 0), data=None
        )

        split_results_task = split_results_task_gen()
        self._task_graph.add_edge(_mitres, split_results_task, key=(0, 0), data=None)
        self._task_graph.add_edge(
            collate_circuit_shots_task, split_results_task, key=(1, 1), data=None
        )

        get_expectations_task = get_expectations_task_gen()
        self._task_graph.add_edge(
            split_results_task, get_expectations_task, key=(0, 0), data=None
        )
        self._task_graph.add_edge(
            filter_observable_tracker_task, get_expectations_task, key=(1, 1), data=None
        )
        self._task_graph.add_edge(get_expectations_task, self._o, key=(0, 0), data=None)

    def check_prepend_wires(self, task: Union[MitTask, TaskGraph]) -> bool:
        """
        Confirms that the number of out wires the passed task has is equal
        to the number of out wires from the input, and that the number
        of in wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[ObservableExperiment] and that its
        return type is Tuple[List[ObservableExperiment]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.

        :return: True if task is suitably for prepending, False if not.
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_out_wires == self.n_in_wires)
            and (task.n_in_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[ObservableExperiment])
            and (sig.return_annotation == Tuple[List[ObservableExperiment]])
        )

    def check_append_wires(self, task: Union[MitTask, TaskGraph]) -> bool:
        """
        Confirms that the number of in wires the passed task has is equal
        to the number of in wires to the output, and that the number
        of out wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[QubitPauliOperator] and that its
        return type is Tuple[List[QubitPauliOperator]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.

        :return: True if task is suitably for appending, False if not.
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_in_wires == self.n_out_wires)
            and (task.n_out_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[QubitPauliOperator])
            and (sig.return_annotation == Tuple[List[QubitPauliOperator]])
        )

    def __str__(self):
        return f"<MitEx::{self._label}>"

    def __call__(  # type: ignore[override]
        self,
        experiment_wires: List[List[ObservableExperiment]],
        cache: bool = False,
        characterisation: dict = {},
    ) -> Tuple[List[QubitPauliOperator]]:
        return cast(
            Tuple[List[QubitPauliOperator]],
            super().run(
                cast(List[Wire], experiment_wires),
                cache=cache,
                characterisation=characterisation,
            ),
        )

    def from_TaskGraph(self, task_graph: TaskGraph):
        """
        Returns a MitEx object from a TaskGraph object.

        :param task_graph: TaskGraph object to copy tasks from.

        :return: Copied TaskGraph as MitEx
        """
        if task_graph.n_in_wires != 1 or task_graph.n_out_wires != 1:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitEx.run type signature. Number of in and out wires does not match."
            )
        # can index as previous check means there should only be one edge
        input_parameters = list(
            inspect.signature(
                list(task_graph._task_graph.out_edges(task_graph._i))[0][1].run
            ).parameters.values()
        )
        if (
            len(input_parameters) != 1
            and input_parameters[0].annotation != List[ObservableExperiment]
        ):
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitEx.run type signature. First MitTask in graph should expect a single argument of List[ObservableExperiment], but expects {}.".format(
                    input_parameters
                )
            )

        # can index as previous check means there should only be one edge
        return_annotation = inspect.signature(
            list(task_graph._task_graph.in_edges(task_graph._o))[0][0].run
        ).return_annotation
        if return_annotation != Tuple[List[QubitPauliOperator]]:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitEx.run type signature. Last MitTask in"
                "task graph should return Tuple[List[QubitPauliOperator]], but returns {}.".format(
                    return_annotation
                )
            )
        self._task_graph = deepcopy(task_graph._task_graph)
        self._label = task_graph._label
        return self

    def parallel(self, task: Union[MitTask, "TaskGraph"]):
        """
        Requests to add new MitTask/TaskGraph to TaskGraph object in parallel.
        Not permitted for MitEx, raises TypeError.

        :param task: New task to be added in parallel.
        """
        raise TypeError("MitEx.parallel forbidden.")

    def add_n_wires(self, num_wires: int):
        """
        Requests to add num_wires number of edges between the input vertex
        and output vertex, with no type restrictions. Not permitted for MitEx,
        raises TypeError.

        :param num_wires: Number of edges to add between input and output vertices.
        """
        raise TypeError("MitEx.add_n_wires forbidden.")

    def add_wire(self):
        """
        Requests to add a single edge between the input vertex and output vertex.
        Not permitted for MitEx, raises TypeError.
        """
        raise TypeError("MitEx.add_wire forbidden.")

    def run(  # type: ignore[override]
        self,
        mitex_wires: List[ObservableExperiment],
        cache: bool = False,
        characterisation: dict = {},
    ) -> List[QubitPauliOperator]:
        """
        Overloaded run method.
        A single observable experiment is defined by a Tuple containg an Ansatz
        Circuit object and an ObservableTracker object.
        An AnsatzCircuit is a tuple containing a Circuit without measures (the ansatz circuit), the number
        of shots to be taken of each Measurement Circuit later run and a SymbolsDict object
        holding a dictionary between Circuit Symbolics (if present) and values
        for substituting them with when running measurement circuits (i.e. parameters).
        It is useful to keep parameters as symbolics until measurement circuits are executed
        as some mitigation methods benefit from being able to run differently parameterised circuits.
        An ObservableTracker object is defined by a QubitPauliOperator defining the observable of interest
        (a dictionary between QubitPauliStrings and coefficients), and later stores MeasurementCircuit
        objects for running on devices.

        :param mitex_wires: Each Tuple pertains to a different Observable measuring experiment, and contains
            the minimum amount of information to run an Mitigated Experiment for calculating observables.

        :return: Observable experiment results as QubitPauliOperator, where values are expectations.
        """
        return self([mitex_wires], cache, characterisation)[0]

    def run_basic(
        self, mitex_wires: List[Tuple[CircuitShots, QubitPauliOperator]]
    ) -> List[QubitPauliOperator]:
        """
        Additional run option for MitEx which simplifies the arguments required for basic experiments.
        An experiment is defined here as just the ansatz circuit, number of shots and the operator being measured.
        This method converts these into the arguments for using run, calls run and then returns the
        desired operator.

        :param mitex_wires: Each tuple pertains to a different basic observable measuring experiment.

        :return: Observable experiment results as QubitPauliOperator, where values are expectations.
        """
        run_wires = [
            ObservableExperiment(
                AnsatzCircuit(m[0].Circuit, m[0].Shots, SymbolsDict()),
                ObservableTracker(m[1]),
            )
            for m in mitex_wires
        ]
        return self.run(run_wires)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from copy import copy
from typing import Dict, List, Tuple, Union, cast

from numpy import mean
from pytket import Circuit
from pytket.backends import Backend
from pytket.pauli import Pauli, QubitPauliString
from pytket.tailoring import apply_clifford_basis_change
from pytket.transform import Transform
from pytket.utils import QubitPauliOperator
from sympy.core.expr import Expr  # type: ignore

from qermit import (
    AnsatzCircuit,
    MeasurementCircuit,
    MitEx,
    MitTask,
    ObservableExperiment,
    ObservableTracker,
    SymbolsDict,
    TaskGraph,
)
from qermit.taskgraph.mitex import gen_compiled_MitRes, get_basic_measurement_circuit


def get_clifford_mcs(input_circuit: Circuit) -> List[MeasurementCircuit]:
    """
    For given Circuit, rebases and substitutes all non-Clifford angles with symbols.
    Then, makes MeasurementCircuit objects wherein each Circuit is the same, but each SymbolsDict
    object varies with some different set of Clifford angles.

    :param input_circuit: Circuit to make all-non Clifford gates paramterised.

    :return: New MeasurementCircuits with Clifford parameters
    """
    copy_circ = input_circuit.copy()
    symbols = copy_circ.free_symbols()
    symbols_dict = dict()
    for s in symbols:
        symbols_dict[s] = cast(Union[None, float], float(0))
    sd = SymbolsDict.symbols_from_dict(symbols_dict)
    return [MeasurementCircuit(copy_circ, sd)]


# TODO: make prepare -1 eigenstate also
def preparation_circuit_for_partition(
    clifford_circuit: Circuit, partition: List[QubitPauliString]
) -> Circuit:
    """
    For each Pauli string in partition, finds Pauli string produced by applying a Clifford basis change from given Clifford circuit.
    Returns a state preparation circuit for preparing a +1 eigenstate of all basis changed Pauli strings.
    """

    eigenstate_circuit = Circuit(0)
    for q in clifford_circuit.qubits:
        eigenstate_circuit.add_qubit(q)
    # to make +1 eigenstate of all transformed strings
    # transform string, then append gates for preparing
    # +1 eigenstate to circuit
    # use RemoveRedundancies after to minimise gates required in cosntruction
    Transform.RebaseToCliffordSingles().apply(clifford_circuit)
    for string in partition:
        transformed_string = apply_clifford_basis_change(string, clifford_circuit)
        transformed_dict = transformed_string.map
        for qubit in transformed_dict:
            if transformed_dict[qubit] == Pauli.X:
                eigenstate_circuit.H(qubit)
            if transformed_dict[qubit] == Pauli.Y:
                eigenstate_circuit.V(qubit)
        Transform.RemoveRedundancies().apply(eigenstate_circuit)
    return eigenstate_circuit


def DFSC_circuit_task_gen() -> MitTask:
    """
    For each experiment, the ansatz circuit has all symbolic gates substituted for Clifford angles (in this case, all 0's).
    If any non symbolic gates are non Clifford, an error is thrown.
    For each Clifford ansatz circuit, a new ObservableTracke is forme with new measurement circuits
    added for each Qubit Pauli String in the operator.

    :return: MitTask object that produces characterisation circuits for DFSC on a new wire as new experiments
    """

    def task(
        obj,
        measurement_wires: List[ObservableExperiment],
    ) -> Tuple[
        List[ObservableExperiment],
        List[List[List[ObservableExperiment]]],
    ]:
        """
        :param measurement_wires: A list of tuples, each tuple representing a different experiment

        :return: Original experiment wires and another list of characterisation experiments for each original experiment.
        These are organised in later task.
        """
        characterisation_wires = []
        for measurement_wire in measurement_wires:
            ansatz_circ = measurement_wire.AnsatzCircuit
            base_circ = ansatz_circ[0]
            tracker = measurement_wire.ObservableTracker
            # Given a circuit, sets all Symbols to Clifford angles
            # If circuit has non-Clifford elements not as symbolics, error thrown
            clifford_circuits = get_clifford_mcs(base_circ)

            # make a new ObservableTracker for holding characterisation circuits
            single_experiment_wires = []
            for c in clifford_circuits:
                clifford_trackers = []
                for string in tracker._qps_to_indices.keys():
                    # each characterisation circuit must have its own observable tracker
                    # this is as the 'ansatz' circuit changes for each string + Clifford combo
                    # as state preparation changes
                    new_tracker = ObservableTracker(QubitPauliOperator({string: 1}))
                    # get measurement circuit
                    measurement_circuit_info = get_basic_measurement_circuit(string)

                    para_circuit = c.get_parametric_circuit()
                    # get preparation circuit for partition, though only pass 1 string
                    prep_circuit = preparation_circuit_for_partition(
                        para_circuit, [string]
                    )
                    # add components to get characterisation circuit
                    prep_circuit.append(para_circuit)
                    new_ansatz_c = AnsatzCircuit(
                        Circuit=prep_circuit.copy(),
                        Shots=ansatz_circ[1],
                        SymbolsDict=c._symbols,
                    )
                    prep_circuit.append(measurement_circuit_info[0])
                    # add to new tracker for given Clifford circuit

                    new_tracker.add_measurement_circuit(
                        MeasurementCircuit(prep_circuit, c._symbols),
                        [measurement_circuit_info[1]],
                    )
                    clifford_trackers.append(
                        ObservableExperiment(
                            AnsatzCircuit=new_ansatz_c, ObservableTracker=new_tracker
                        )
                    )
                # a single experiment being all pauli string measurement circuits for a single Clifford Circuit
                single_experiment_wires.append(clifford_trackers)
            # single experiment wires being all observable trackers for all sampled Clifford circuits with
            # added measurement circuits for all pauli strings in given experiment
            characterisation_wires.append(single_experiment_wires)
        return (measurement_wires, characterisation_wires)

    return MitTask(_label="DFSCCircuits", _n_in_wires=1, _n_out_wires=2, _method=task)


def DFSC_collater_task_gen() -> MitTask:
    """
    For each experiment passed to MitEx, DFSC characterisation produces an ObservableTracker
    of a single Measurement Circuit for each combination of Clifford circuit produced, eigenstates preparation
    and QubitPauliString in operator, via several nested Lists.
    This task unpackages these Lists into a single List as suitable for input to MitEx objects.
    It also stores information required to produce characterisation from resulting QubitPauliOperators out of
    MitEx object.

    :return: MitTask object that collates many BackendResult objects for a single
        frame randomisation instance and converts them into a single
        BackendResult object.
    """

    def task(
        obj,
        all_characterisation_trackers: List[List[List[ObservableExperiment]]],
    ) -> Tuple[List[ObservableExperiment], List[int]]:
        """
        :param all_characterisation_trackers: Experiment wires; outer list is experiments, second outer list
        is Cliffords, inner list is qubit pauli strings.

        :return: Wire 1; All individual experiments collated into a single wire.
        Wire 2; Indexing to produce characterisation later.
        """
        organisation_indices = []
        collated_experiments = []
        for experiment_char in all_characterisation_trackers:
            # individual list is for some cliffords
            # qps is stored in output
            base_len = 0
            for cliff_ots in experiment_char:
                base_len += len(cliff_ots)
                for qps_ot in cliff_ots:
                    collated_experiments.append(qps_ot)
            organisation_indices.append(base_len)
        return (collated_experiments, organisation_indices)

    return MitTask(_label="DFSCCollation", _n_in_wires=1, _n_out_wires=2, _method=task)


def DFSC_characterisation_task_gen() -> MitTask:
    """
    Given characterisation results for all experiments, Clifford circuits and QubitPauliStrings, produces
    a characterisation result for each Experiment.

    :return: MitTask object for organising and calculating characterisation.
    """

    def task(
        obj,
        characterisation_results: List[QubitPauliOperator],
        experiment_indexing: List[int],
    ) -> Tuple[List[QubitPauliOperator]]:
        """
        :param characterisation_results: All QubitPauliOperators returned from running experiment through some MitEx object
        :param experiment_indexing: Number of characteriastion results for each experiment, used to split results up.

        :return: Collated characterisation results, one QubitPauliOperator characterisation for each experiment
        """
        split_results = []
        lower_bound = 0
        for size in experiment_indexing:
            upper_bound = lower_bound + size
            split_results.append(characterisation_results[lower_bound:upper_bound])
            lower_bound = upper_bound

        characterisation_qpos = []
        for experiment_results in split_results:
            characterisation_dict: Dict[QubitPauliString, List[float]] = dict()
            # add all expectations for each Clifford + String combo to dict
            for qpo in experiment_results:
                for string in qpo._dict:
                    if string not in characterisation_dict:
                        characterisation_dict[string] = list()
                    characterisation_dict[string].append(float(qpo._dict[string]))
            # set entry to average of values in list
            # add characterisation for DFSC to output list
            characterisation_qpos.append(
                QubitPauliOperator(
                    {
                        k: cast(Union[complex, Expr], mean(characterisation_dict[k]))
                        for k in characterisation_dict
                    }
                )
            )
        # number of characterisation qpos should match original number of experiments
        return (characterisation_qpos,)

    return MitTask(
        _label="DFSCCharacterisation", _n_in_wires=2, _n_out_wires=1, _method=task
    )


def DFSC_correction_task_gen(zero_threshold: float) -> MitTask:
    """
    For each experiment expectation, if characterisation value greater than threshold, divide experiment expectation
    by characterisation value to correct for depolarising noise.

    :param zero_threshold: Method does not correct for zero characterisation expectation values, threshold for this zero limit.

    :return: Function for DFSC correction.
    """

    def task(
        obj,
        experiment_results: List[QubitPauliOperator],
        characterisation_results: List[QubitPauliOperator],
    ) -> Tuple[List[QubitPauliOperator]]:
        """
        :param experiment_results: QubitPauliOperators corresponding to expectations for all observable experiments.
        :param characteriastion_results: QubitPauliOperators corresponding to expectations for all characterisation experiments.

        :return: Corrected expectations as QubitPauliOperator objects.
        """
        if len(experiment_results) != len(characterisation_results):
            raise ValueError(
                "{} Experiment results and {} Characterisation results: mismatch for DFSC correction.".format(
                    len(experiment_results), len(characterisation_results)
                ),
            )
        corrected_results = []
        for experiment_qpo, characterisation_qpo in zip(
            experiment_results, characterisation_results
        ):
            new_qpo = dict()
            for key in experiment_qpo._dict:
                val = characterisation_qpo._dict[key]
                if val > zero_threshold:
                    new_qpo[key] = experiment_qpo._dict[key] / val
                else:
                    new_qpo[key] = experiment_qpo._dict[key]
            corrected_results.append(QubitPauliOperator(new_qpo))
        return (corrected_results,)

    return MitTask(_label="DFSCCorrection", _n_in_wires=2, _n_out_wires=1, _method=task)


def gen_DFSC_MitEx(backend: Backend, **kwargs) -> MitEx:
    """
    Produces a MitEx object that applies DFSC characterisation to all experiment results.

    :param backend: Backend experiments are run through.
    :key experiment_mitex: MitEx object observable experiments are run through
    :key characterisation_mitex: MitEX object characteriastion experiments are run through.

    :return: MitEx object for automatic DFSC correction of circuits.
    """

    _experiment_mitex = copy(
        kwargs.get(
            "experiment_mitex",
            MitEx(
                backend,
                _label="ExperimentMitex",
                mitres=gen_compiled_MitRes(backend, 0),
            ),
        )
    )
    _characterisation_mitex = copy(
        kwargs.get(
            "characterisation_mitex",
            MitEx(
                backend,
                _label="CharacterisationMitex",
                mitres=gen_compiled_MitRes(backend, 0),
            ),
        )
    )

    _characterisation_taskgraph = TaskGraph().from_TaskGraph(_characterisation_mitex)
    _experiment_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitex)

    _characterisation_taskgraph.add_wire()
    _characterisation_taskgraph.prepend(DFSC_collater_task_gen())
    _characterisation_taskgraph.append(DFSC_characterisation_task_gen())

    _experiment_taskgraph.parallel(
        MitEx(backend).from_TaskGraph(_characterisation_taskgraph)
    )
    _experiment_taskgraph.prepend(DFSC_circuit_task_gen())
    _experiment_taskgraph.append(
        DFSC_correction_task_gen(kwargs.get("DFSC_threshold", 0.01))
    )
    return MitEx(backend).from_TaskGraph(_experiment_taskgraph)



from typing import List, Tuple

from pytket.backends import Backend
from pytket.passes import DecomposeBoxes

from qermit import CircuitShots, MitRes, MitTask, TaskGraph
from qermit.coherent_pauli_checks.box_clifford_subcircuits import BoxClifford
from qermit.postselection.postselect_manager import PostselectMgr
from qermit.postselection.postselect_mitres import gen_postselect_task

from .pauli_sampler import PauliSampler


def gen_find_cliffords_task() -> MitTask:
    """Generator for MitTask for finding Clifford subcircuits within
    given circuits. Inputs and outputs are CircuitShots. Output circuits
    have Clifford sub circuits wrapped in CircBox.

    :return: Task for finding Clifford subcircuits.
    """

    def task(_, circ_shots_list: List[CircuitShots]) -> Tuple[List[CircuitShots]]:
        """Function wrapping Clifford subcircuits in CircBox.

        :param circ_shots_list: List of circuit shots.
        :return: Identical circuits with Clifford circuit wrapped
            in CircBox.
        """
        cliff_circ_shots_list = []

        for circ_shots in circ_shots_list:
            cliff_circ = circ_shots.Circuit.copy()
            BoxClifford().apply(cliff_circ)
            cliff_circ_shots_list.append(
                CircuitShots(
                    Circuit=cliff_circ,
                    Shots=circ_shots.Shots,
                )
            )

        return (cliff_circ_shots_list,)

    return MitTask(
        _label="FindCliffordSubcircuits",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=task,
    )


def gen_check_circuit_task(pauli_sampler: PauliSampler) -> MitTask:
    """Generator for MitTask applying Pauli checks to circuits.

    :param pauli_sampler: PauliSampler to use to sample checks.
    :return: Task for adding Pauli checks.
    """

    def task(
        _, circ_shots_list: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[PostselectMgr]]:
        """Task for adding Pauli checks to given circuits.

        :param circ_shots_list: Circuits to add Pauli checks to.
        :return: Circuits with pauli checks added.
        """
        checked_circ_shots_list = []
        postselect_mgr_list = []

        for circ_shots in circ_shots_list:
            checked_circuit, postselect_cbits = (
                pauli_sampler.add_pauli_checks_to_circbox(circuit=circ_shots.Circuit)
            )

            postselect_mgr_list.append(
                PostselectMgr(
                    compute_cbits=circ_shots.Circuit.bits,
                    postselect_cbits=list(postselect_cbits),
                )
            )

            DecomposeBoxes().apply(checked_circuit)

            checked_circ_shots_list.append(
                CircuitShots(
                    Circuit=checked_circuit,
                    Shots=circ_shots.Shots,
                )
            )

        return (checked_circ_shots_list, postselect_mgr_list)

    return MitTask(
        _label="CheckCircuits",
        _n_in_wires=1,
        _n_out_wires=2,
        _method=task,
    )


def gen_coherent_pauli_check_mitres(
    backend: Backend,
    pauli_sampler: PauliSampler,
) -> MitRes:
    """Generator for MitRes performing mitigation through
    Coherent Pauli Checks.

    :param backend: Backend to perform perform experiment with.
    :param pauli_sampler: Sampler to use to apply Pauli checks.
    :return: MitRes performing mitigation through
    Coherent Pauli Checks.
    """
    _mitres = MitRes(backend, _label="PostselectionMitRes")
    _taskgraph = TaskGraph().from_TaskGraph(_mitres)

    _taskgraph.add_wire()

    _taskgraph.prepend(gen_check_circuit_task(pauli_sampler))
    _taskgraph.prepend(gen_find_cliffords_task())
    _taskgraph.append(gen_postselect_task())

    return MitRes(backend).from_TaskGraph(_taskgraph)



import multiprocessing
import uuid
from collections import Counter
from typing import Dict, Iterable, Iterator, List, Optional, Sequence, Tuple

from pytket import Bit, Circuit
from pytket.backends.backend import Backend
from pytket.backends.backendresult import BackendResult
from pytket.backends.resulthandle import ResultHandle
from pytket.extensions.qiskit import AerBackend
from pytket.passes import BasePass, CustomPass
from pytket.utils.outcomearray import OutcomeArray


class TranspilerBackend:
    """
    Provides a backend like interface for noise simulation via compiler passes.
    In particular, for each shot a new circuit is generated by applying the
    given compiler pass.

    Attributes:
        transpiler: Compiler pass to apply to simulate noise on a single instance of a circuit.
        max_batch_size: Shots are simulated in batches. This is the largest shot batch size permitted.
        result_dict: A dictionary mapping handles to results.
        backend: Backend used to simulate compiled circuits.
        n_cores: The number of cores used when simulating shots in parallel.
    """

    transpiler: BasePass
    max_batch_size: int
    result_dict: Dict[ResultHandle, BackendResult]
    n_cores: int

    def __init__(
        self,
        transpiler: BasePass,
        result_dict: Dict[ResultHandle, BackendResult] = {},
        max_batch_size: int = 1000,
        n_cores: int = 1,
        backend: Backend = AerBackend(),
    ):
        """Initialisation method.

        :param transpiler: Compiler to use during noise simulation.
        :param max_batch_size: Size of the largest batch of shots,
            defaults to 1000. The total number of shots is distributed between
            bathes of size 1000 plus a smaller batch for left over shot.
            These batches will be distributed to multiple cores.
        :param result_dict: Results dictionary, may be used to store existing
            results within backend, defaults to {}
        :param n_cores: Shots will be taken in parallel. This parameter
            specifies the number of cores to use. The default is to use
            one core.
        """

        self.transpiler = transpiler

        self.max_batch_size = max_batch_size
        self.result_dict = result_dict
        self.n_cores = n_cores
        self.backend = backend

    @property
    def _uses_lightsabre(self) -> bool:
        # lightsaber is a routing method used by pytket-qiskit
        return False

    def default_compilation_pass(self, **kwargs) -> BasePass:
        """Return a compiler pass which has no affect on the circuit."""
        return CustomPass(transform=lambda circuit: circuit)

    def rebase_pass(self) -> BasePass:
        """Return a compiler pass which has no affect on the circuit."""
        return CustomPass(transform=lambda circuit: circuit)

    def run_circuit(
        self,
        circuit: Circuit,
        n_shots: int,
        **kwargs,
    ) -> BackendResult:
        """Return results of running one circuit.

        :param circuit: Circuit to run
        :param n_shots: Number of shots to be taken from circuit.
        :return: Result of running circuit.
        """

        handle = self.process_circuit(circuit, n_shots, **kwargs)
        return self.get_result(handle=handle)

    def process_circuits(
        self,
        circuits: Sequence[Circuit],
        n_shots: Sequence[int],
    ) -> List[ResultHandle]:
        """Processes a collection of circuits by making use multiple calls
        to process_circuit.

        :param circuits: A collection of circuit to run.
        :param n_shots: The number of shots which should be taken from
            each circuit.
        :return: The result handle for each circuit.
        """

        return [
            self.process_circuit(circuit=circuit, n_shots=n)
            for circuit, n in zip(circuits, n_shots)
        ]

    def process_circuit(
        self,
        circuit: Circuit,
        n_shots: int,
        **kwargs,
    ) -> ResultHandle:
        """[summary]

        :param circuit: Submits circuit to run on noisy backend.
        :param n_shots: Number of shots to take from circuit.
        :return: Handle identifying results in `result_dict`.
        """

        handle = ResultHandle(str(uuid.uuid4()))

        counts = self.get_counts(
            circuit=circuit,
            n_shots=n_shots,
            cbits=circuit.bits,
        )

        self.result_dict[handle] = BackendResult(
            counts=Counter(
                {OutcomeArray.from_readouts([key]): val for key, val in counts.items()}
            ),
            c_bits=circuit.bits,
        )

        return handle

    def get_results(self, handles: Iterable[ResultHandle]) -> List[BackendResult]:
        """Get the results corresponding to a collection of result handles.

        :param handles: A collection of handles to retrieve.
        :return: The results corresponding to the given collection of
        """
        return [self.get_result(handle) for handle in handles]

    def get_result(self, handle: ResultHandle) -> BackendResult:
        """Retrieve result from backend.

        :param handle: Handle identifying result.
        :return: Result corresponding to handle.
        """
        return self.result_dict[handle]

    def _gen_transpiled_circuit(self, circuit: Circuit) -> Circuit:
        """Generate compiled circuit by copying and compiling it.

        :param circuit: Circuit to be compiled.
        :return: Compiled circuit.
        """

        transpiled_circuit = circuit.copy()
        self.transpiler.apply(transpiled_circuit)
        self.backend.rebase_pass().apply(transpiled_circuit)
        return transpiled_circuit

    def _gen_batches(self, circuit: Circuit, n_shots: int) -> Iterator[List[Circuit]]:
        """Iterator generating lists of circuits of size `max_batch_size`
            until all shots have been accounted for.

        :param circuit: Circuit to batch into shots.
        :param n_shots: Number of shots to take from circuit.
        :return: List of compiled circuits, which is to say noisy circuits.
        """

        # Return lists of size max_batch_size containing unique
        # compiled instances of the given circuit.
        for _ in range(n_shots // self.max_batch_size):
            yield [
                self._gen_transpiled_circuit(circuit)
                for _ in range(self.max_batch_size)
            ]

        # If less than max_batch_size shots remain to be returned,
        # return what's left.
        if n_shots % self.max_batch_size > 0:
            yield [
                self._gen_transpiled_circuit(circuit)
                for _ in range(n_shots % self.max_batch_size)
            ]

    @staticmethod
    def _get_batch_counts(
        circuit_list: List[Circuit],
        cbits_list: Optional[List[List]],
    ) -> Counter[Tuple[int, ...]]:
        """Run each circuit in the given list for one shot,
        collating the results into a single counter.

        :param circuit_list: The list of circuits to run for one shot each.
        :param cbits_list: The classical bits to return the measurements of
        :return: The collated counter object.
        """

        if cbits_list is not None:
            cbits = [Bit.from_list(cbit_list) for cbit_list in cbits_list]
        else:
            cbits = None

        backend = AerBackend()

        result_list = backend.run_circuits(circuit_list, n_shots=1)
        return sum(
            (result.get_counts(cbits=cbits) for result in result_list), Counter()
        )

    def get_counts(
        self,
        circuit: Circuit,
        n_shots: int,
        cbits: Optional[List[Bit]] = None,
    ) -> Counter[Tuple[int, ...]]:
        """Generate shots from the given circuit.

        :param circuit: Circuit to take shots from.
        :param n_shots: Number of shots to take from circuit.
        :param cbits: Classical bits to return shots from,
            defaults to returning all.
        :return: Counter detailing shots from circuit.
        """

        if self.n_cores > 1:
            if cbits is not None:
                cbits_list = [cbit.to_list() for cbit in cbits]
            else:
                cbits_list = None

            with multiprocessing.Pool(self.n_cores) as pool:
                processes = [
                    pool.apply_async(
                        self._get_batch_counts, args=(circuit_list, cbits_list)
                    )
                    for circuit_list in self._gen_batches(circuit, n_shots)
                ]
                counter_list = [p.get() for p in processes]

            return sum(counter_list, Counter())

        else:
            counter: Counter = Counter()
            for circuit_list in self._gen_batches(circuit, n_shots):
                result_list = self.backend.run_circuits(circuit_list, n_shots=1)
                counter += sum(
                    (result.get_counts(cbits=cbits) for result in result_list),
                    Counter(),
                )

            return counter



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from copy import deepcopy
from itertools import repeat
from typing import List, OrderedDict, Sequence, Tuple, Union, cast

import networkx as nx  # type: ignore
import numpy as np
from pytket import Bit, Circuit
from pytket.backends import Backend, ResultHandle
from pytket.backends.backendresult import BackendResult
from pytket.utils.outcomearray import OutcomeArray

from .mittask import (
    CircuitShots,
    IOTask,
    MitTask,
    Wire,
)
from .task_graph import TaskGraph


def backend_compile_circuit_shots_task_gen(
    backend: Backend, optimisation_level: int = 1
) -> MitTask:
    """
    For each circuit in passed List[CircuitShots] argument, pass
    circuit to compile_circuit method of given backend with given optimisation level.
    Returns new wire wherein each circuit has been compiled.

    :param backend: Backend object from which compile circuit method is called
    :param optimisation_level: Optimisation level of backend called.
    """

    def task(obj, circ_shots: List[CircuitShots]) -> Tuple[List[CircuitShots]]:
        return (
            [
                CircuitShots(
                    backend.get_compiled_circuit(
                        cs.Circuit, optimisation_level=optimisation_level
                    ),
                    cs.Shots,
                )
                for cs in circ_shots
            ],
        )

    return MitTask(
        _label="CompileCircuitShots", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def backend_handle_task_gen(backend: Backend) -> MitTask:
    """
    Passes every tuple of Circuit and Shots to the backend object MitTask
    is defined by, returning a handle for later retrieving results for each circuit.
    If different numbers of shots are passed, each circuit is run with the maximum number of shots.

    :param backend: Backend circuits are run through.
    :return: Pure function that adds passes circuits to backend and gets handles.
    """

    def task(obj, circuit_wires: List[CircuitShots]) -> Tuple[List[ResultHandle]]:
        """
        :param circuit_wires: Circuits to be run on backend, number of shots to run of each.

        :return: ResultHandles from process_circuits method.
        """

        if len(circuit_wires) != 0:
            circs, shots = map(list, zip(*circuit_wires))

            results = backend.process_circuits(
                cast(List[Circuit], circs), n_shots=cast(Sequence[int], shots)
            )

            return (results,)
        else:
            return ([],)

    return MitTask(
        _label="CircuitsToHandles", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def backend_res_task_gen(backend: Backend) -> MitTask:
    """
    For each ResultHandle passed to task, retrieves a BackendResult object from
    the backend the task is defined by.

    :param backend: backend holding results for handles.
    """

    def task(obj, handles: List[ResultHandle]) -> Tuple[List[BackendResult]]:
        """
        :param handles: ResultHandle objects previously produced from backend.

        :return: For each ResultHandle in handles, a BackendResult object retrieved from backend.
        """
        results = backend.get_results(handles)

        return (results,)

    return MitTask(
        _label="HandlesToResults", _n_in_wires=1, _n_out_wires=1, _method=task
    )


class MitRes(TaskGraph):
    """
    A TaskGraph extension of mitigation of counts/shots for individual circuits.
    """

    def __init__(
        self,
        backend: Backend,
        _label: str = "MitRes",
    ) -> None:
        """
        MitRes objects are defined by the backend objects all circuits are executed on.


        :param backend: Pytket backend default constructor which tasks are generated from.
        :param label: Name for identification of MitRes object.
        """
        # set member variables
        self._label = _label
        self.G = None
        self.characterisation: dict = {}

        # default constructor runs all circuits through passed Backend
        self._task_graph = nx.MultiDiGraph()

        # if requested, all data is held in cache and can be accessed after running
        self._cache: OrderedDict[str, Tuple[MitTask, List[Wire]]] = OrderedDict()

        c2h = backend_handle_task_gen(backend)
        h2r = backend_res_task_gen(backend)

        self._i, self._o = IOTask.Input, IOTask.Output
        self._task_graph.add_edge(self._i, c2h, key=(0, 0), data=None)
        self._task_graph.add_edge(c2h, h2r, key=(0, 0), data=None)
        self._task_graph.add_edge(h2r, self._o, key=(0, 0), data=None)

    def check_prepend_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of out wires the passed task has is equal
        to the number of out wires from the input, and that the number
        of in wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[CircuitShots] and that its
        return type is Tuple[List[CircuitShots]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.

        :return: True if task is suitable for prepending, False if not.
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_out_wires == self.n_in_wires)
            and (task.n_in_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[CircuitShots])
            and (sig.return_annotation == Tuple[List[CircuitShots]])
        )

    def check_append_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of in wires the passed task has is equal
        to the number of in wires to the output, and that the number
        of out wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[BackendResult] and that its
        return type is Tuple[List[BackendResult]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.

        :return: True if task is suitable for apppending, False if not.
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_in_wires == self.n_out_wires)
            and (task.n_out_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[BackendResult])
            and (sig.return_annotation == Tuple[List[BackendResult]])
        )

    def __str__(self) -> str:
        return f"<MitRes::{self._label}>"

    def __call__(  # type: ignore[override]
        self,
        circuits_wire: List[List[CircuitShots]],
        cache: bool = False,
        characterisation: dict = {},
    ) -> Tuple[List[BackendResult]]:
        return cast(
            Tuple[List[BackendResult]],
            super().run(
                cast(List[Wire], circuits_wire),
                cache=cache,
                characterisation=characterisation,
            ),
        )

    def from_TaskGraph(self, task_graph: TaskGraph):
        """
        Returns a MitRes object from a TaskGraph object.

        :param task_graph: TaskGraph object to copy tasks from.
        :return: Copied TaskGraph as MitRes
        """
        if task_graph.n_in_wires != 1 or task_graph.n_out_wires != 1:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRun.run type signature. Number of in and out wires does not match."
            )
        # can index as previous check means there should only be one edge
        input_parameters = list(
            inspect.signature(
                list(task_graph._task_graph.out_edges(task_graph._i))[0][1].run
            ).parameters.values()
        )
        if (
            len(input_parameters) != 1
            and input_parameters[0].annotation != List[CircuitShots]
        ):
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRes.run type signature. First MitTask in graph should expect a single argument of List[Circuitshots], but expects {}.".format(
                    input_parameters
                )
            )

        # can index as previous check means there should only be one edge
        return_annotation = inspect.signature(
            list(task_graph._task_graph.in_edges(task_graph._o))[0][0].run
        ).return_annotation
        if return_annotation != Tuple[List[BackendResult]]:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRes.run type signature. Last MitTask in"
                "task graph should return Tuple[List[BackendResult]], but returns {}.".format(
                    return_annotation
                )
            )
        self._task_graph = deepcopy(task_graph._task_graph)
        self._label = task_graph._label
        return self

    def parallel(self, task: Union[MitTask, "TaskGraph"]):
        """
        Requests to add new MitTask/TaskGraph to TaskGraph object in parallel.
        Not permitted for MitRes, raises TypeError.

        :param task: New task to be added in parallel.
        """
        raise TypeError("MitRes.parallel forbidden.")

    def add_n_wires(self, num_wires: int):
        """
        Requests to add num_wires number of edges between the input vertex
        and output vertex, with no type restrictions. Not permitted for MitRes,
        raises TypeError.


        :param num_wires: Number of edges to add between input and output vertices.
        """
        raise TypeError("MitRes.add_n_wires forbidden.")

    def add_wire(self):
        """
        Requests to add a single edge between the input vertex and output vertex.
        Not permitted for MitRes, raises TypeError.
        """
        raise TypeError("MitRes.add_wire forbidden.")

    def run(  # type: ignore[override]
        self,
        circuit_shots: List[CircuitShots],
        cache: bool = False,
        characterisation: dict = {},
    ) -> List[BackendResult]:
        """
        Overloaded run method from TaskGraph class to add type checking.
        A single experiment is defined by a Tuple containing a circuit to be run
        on some backend and the number of shots to take of said circuit.
        For each combination of Circuit and shots run returns a BackendResult object
        containing counts/shots.

        :param circuit_shotss: Each tuple in circuit_wires contains a Circuit to run on
            internal backends and the number of shots to take of said circuit.

        :return: A BackendResult object for each combination of circuit and shots.
        """
        return self([circuit_shots], cache, characterisation)[0]


def split_shots_task_gen(max_shots: int) -> MitTask:
    """
    When the number of shots requested is higher than max_shots, this task
    splits the jobs into several individual jobs, where the total number of
    shots matches that initially requested.

    :param max_shots: The maximum number of shots per job.
    :return: A task dividing the job into several smaller ones.
    """

    def task(
        obj, circuit_wires: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[int]]:
        """
        This task divides a job into several smaller ones, if the number of
        shots requested is larger then the maximum allowed.

        :param circuit_wires: A list of the circuits to be run, and the number
        of shots to be taken.
        :return: A new list of circuits, some of which repeat with a smaller
        number of shots.
        """

        # A new list of circuits and shots
        split_circuit_wires = []
        # and index of which circuits in the new list correspond to which in
        # the original list. This is to say all circuits in the new list
        # assigned an index i by this index list correspond to original
        # circuit i.
        split_index: List[int] = []

        # For each circuit, check if the number of shots requested is greater
        # then the maximum, and divide the job into several smaller ones if
        # this is the case.
        for i, circ_shots in enumerate(circuit_wires):
            div_val = circ_shots.Shots // max_shots
            # Divide into jobs of size max_shots
            if circ_shots.Shots > max_shots:
                split_circuit_wires.extend(
                    [
                        CircuitShots(circ_shots.Circuit, max_shots)
                        for _ in range(div_val)
                    ]
                )
            # Add remaining shots
            if circ_shots.Shots % max_shots > 0:
                split_circuit_wires.append(
                    CircuitShots(circ_shots.Circuit, circ_shots.Shots % max_shots)
                )
            # Add new job indexes
            split_index.extend(
                repeat(i, div_val + min(1, circ_shots.Shots % max_shots))
            )

        return (split_circuit_wires, split_index)

    return MitTask(_label="SplitShots", _n_in_wires=1, _n_out_wires=2, _method=task)


def group_shots_task_gen() -> MitTask:
    """
    Returns task which groups together jobs that correspond to the same
    circuit. Original larger job could have been split by task returned
    by split_shots_task_gen.

    :return: Task merging jobs which correspond to the same circuit.
    """

    def task(
        obj, results_wires: List[BackendResult], split_index: List[int]
    ) -> Tuple[List[BackendResult]]:
        """
        Task which merges jobs which correspond to the same circuit.

        :param results_wires: A list of results, some of which correspond to
        the same circuit.
        :param split_index: A list of indexes identifying which results
        correspond to which circuit. Those results with index i correspond to
        the circuit i.
        :return: A new list of results, where results which correspond to the
        same circuit have been grouped together.
        """

        # A new list of results where some results have been merged.
        merged_results = []

        # A list of lists. Each internal list contains results that correspond
        # to the circuit of the same index.
        grouped_results = [
            [result for result, i in zip(results_wires, split_index) if i == j]
            for j in range(split_index[-1] + 1)
        ]

        # For each circuit, combine the results.
        for chunk in grouped_results:
            # Concatenate shots from all results
            combined_shots = np.concatenate([result.get_shots() for result in chunk])
            outcome_array = OutcomeArray.from_readouts(combined_shots)
            merged_results.append(
                BackendResult(
                    shots=outcome_array, c_bits=cast(Sequence[Bit], chunk[0].c_bits)
                )
            )

        return (merged_results,)

    return MitTask(_label="MergeShots", _n_in_wires=2, _n_out_wires=1, _method=task)


def gen_shot_split_MitRes(backend: Backend, max_shots: int) -> MitRes:
    """
    Wraps a mitres in tasks which ensures shots requested by individual
    jobs never exceed max_shots. It does so by splitting the shots over several
    jobs if necessary.

    :param backend: The backend on which to run the jobs.
    :param max_shots: The maximum number of shots that each job should request.
    :return: A MitRes object.
    """

    _experiment_mitres = MitRes(backend=backend)

    _experiment_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitres)

    _experiment_taskgraph.add_wire()

    # Prepend splitting task.
    split_shots = split_shots_task_gen(max_shots)
    _experiment_taskgraph.prepend(split_shots)

    # Append merging task.
    merge_shots = group_shots_task_gen()
    _experiment_taskgraph.append(merge_shots)

    return MitRes(backend).from_TaskGraph(_experiment_taskgraph)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy, deepcopy
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, List, OrderedDict, Tuple, Union, cast

import networkx as nx  # type: ignore

from .graphviz import _taskgraph_to_graphviz
from .mittask import (
    IOTask,
    MitTask,
    Wire,
)

if TYPE_CHECKING:
    import graphviz as gv  # type: ignore


class TaskGraph:
    """
    The TaskGraph class stores a networkx graph where vertices
    are pure functions or tasks, and edges hold data.
    In the TaskGraph class these tasks and edges have no
    type restrictions, though for the run method to be succesful, the
    types of ports edges are attached to must match.

    :param _label: Name for identification of TaskGraph object.
    """

    def __init__(
        self,
        _label: str = "TaskGraph",
    ) -> None:
        # set member variables
        self._label = _label
        self.G = None
        self.characterisation: dict = {}
        # default constructor runs all circuits through passed Backend
        self._task_graph = nx.MultiDiGraph()

        self._i, self._o = IOTask.Input, IOTask.Output
        self._task_graph.add_edge(self._i, self._o, key=(0, 0), data=None)

        # if requested, all data is held in cache and can be accessed after running
        self._cache: OrderedDict[str, Tuple[MitTask, List[Wire]]] = OrderedDict()

    def from_TaskGraph(self, task_graph: "TaskGraph"):
        """
        Returns a new TaskGraph object from another TaskGraph object.

        :param task_graph: TaskGraph object to copy tasks from.

        :return: Copied TaskGraph
        """
        self._task_graph = deepcopy(task_graph._task_graph)
        self._label = task_graph._label
        self.characterisation = task_graph.characterisation
        return self

    @property
    def tasks(self) -> List[MitTask]:
        """
        Returns a list of all tasks with both input and output ports
        in the TaskGraph.
        """
        return list(self._task_graph)[2:]

    def __call__(self, input_wires: List[Wire]) -> Tuple[List[Wire]]:
        return self.run(input_wires)

    @property
    def label(self) -> str:
        return self._label

    def get_characterisation(self) -> dict:
        return self.characterisation

    def update_characterisation(self, characterisation: dict):
        self.characterisation.update(characterisation)

    def set_characterisation(self, characterisation: dict):
        self.characterisation = characterisation

    @property
    def n_in_wires(self) -> int:
        """
        The number of in wires to a TaskGraph object is defined as the number
        of out edges from the Input Vertex, as when called, a TaskGraph object
        calls the run method which stores input arguments as data on Input vertex
        output edges.
        """
        return len(self._task_graph.out_edges(self._i))

    @property
    def n_out_wires(self) -> int:
        """
        The number of out wires to a TaskGraph object is defined as the number
        of in edges to the Input Vertex, as when called, a TaskGraph object
        calls the run method which after running all tasks, returns
        the data on input edges to the Output Vertex as a tuple.
        """
        return len(self._task_graph.in_edges(self._o))

    def check_prepend_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of out wires of the proposed task to prepend to the
        internal task_graph attribute matches the number of in wires to the graph.

        :param task: Wrapped pure function to prepend to graph

        :return: True if prepend permitted
        """
        return task.n_out_wires == self.n_in_wires

    def check_append_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of in wires of the proposed task to append to the
        internal task_graph attribute matches the number of out wires to the graph.

        :param task: Wrapped pure function to append to graph

        :return: True if append permitted
        """
        return task.n_in_wires == self.n_out_wires

    def __str__(self):
        return f"<TaskGraph::{self._label}>"

    def __repr__(self):
        return str(self)

    def add_n_wires(self, num_wires: int):
        """
        Adds num_wires number of edges between the input vertex
        and output vertex, with no type restrictions.

        :param num_wires: Number of edges to add between input and output vertices.
        """
        for _ in range(num_wires):
            in_port = len(self._task_graph.out_edges(self._i, data=True))
            out_port = len(self._task_graph.in_edges(self._o, data=True))
            self._task_graph.add_edge(
                self._i, self._o, key=(in_port, out_port), data=None
            )

    def add_wire(self):
        """
        Adds a single edge between the input vertex and output vertex.
        """
        self.add_n_wires(1)

    # Add news task to start of TaskGraph
    def prepend(self, task: Union[MitTask, "TaskGraph"]):
        """
        Inserts new task to the start of TaskGraph._task_graph.
        All out edges from the Input vertex are wired as out edges from the task in the same port ordering (types must match).
        New edges also added from the Input vertex to the task (any type permitted), ports ordered in arguments order.

        :param task: New task to be prepended.
        """
        assert self.check_prepend_wires(task)
        # It's possible a single generated MitTask object could be used in different TaskGraph objects
        # via prepend which may lead to a task address expecting input wires from different graphs
        # use of copy here prevents this and task graph generation is not the bottleneck in running mitigation
        # schemes so fine
        task_copy = copy(task)

        for i, edge in enumerate(list(self._task_graph.out_edges(self._i, keys=True))):
            self._task_graph.add_edge(
                task_copy, edge[1], key=(i, edge[2][1]), data=None
            )
            self._task_graph.remove_edge(edge[0], edge[1])

        for port in range(task_copy.n_in_wires):
            self._task_graph.add_edge(self._i, task_copy, key=(port, port), data=None)

    def append(self, task: Union[MitTask, "TaskGraph"]):
        """
        Inserts new task to end of TaskGraph._task_graph.
        All in edges to Output vertex are wired as in edges to task in same port ordering (types must match).
        New edges added from task to Output vertex (any type permitted), ports ordered in arguments order.

        :param task: New task to be appended.
        """
        assert self.check_append_wires(task)
        # It's possible a single generated MitTask object could be used in different TaskGraph objects
        # via append which may lead to a task address expecting input wires from different graphs
        # use of copy here prevents this and task graph generation is not the bottleneck in running mitigation
        # schemes so fine
        task_copy = copy(task)
        for edge in list(self._task_graph.in_edges(self._o, keys=True)):
            self._task_graph.add_edge(edge[0], task_copy, key=edge[2], data=None)

            self._task_graph.remove_edge(edge[0], edge[1])

        for port in range(task_copy.n_out_wires):
            self._task_graph.add_edge(task_copy, self._o, key=(port, port), data=None)

    def decompose_TaskGraph_nodes(self):
        """
        For each node in self._task_graph, if node is a TaskGraph object, substitutes that node
        with the _task_graph structure held inside the node.
        """
        check_for_decompose = True
        while check_for_decompose:
            # get all nodes and iterate through them
            check_for_decompose = False
            node_list = list(nx.topological_sort(self._task_graph))
            for task in node_list:
                # => TaskGraph object with _task_graph attribute for decomposition
                if hasattr(task, "_task_graph"):
                    # relabel task names for ease of viewing wit visualisation methods
                    for sub_task in list(task._task_graph.nodes):
                        # in practice only IOTask
                        if hasattr(sub_task, "_label"):
                            sub_task._label = task._label + sub_task._label

                    task_in_edges = list(self._task_graph.in_edges(task, keys=True))
                    task_out_edges = list(self._task_graph.out_edges(task, keys=True))

                    task_input_out_edges = list(
                        task._task_graph.out_edges(task._i, keys=True)
                    )
                    task_output_in_edges = list(
                        task._task_graph.in_edges(task._o, keys=True)
                    )

                    if (
                        len(
                            set(task_input_out_edges).intersection(
                                set(task_output_in_edges)
                            )
                        )
                        > 0
                    ):
                        raise ValueError(
                            "Decomposition of TaskGraph node {}, not permitted: TaskGraph to be decomposed has edge between input and output vertices.".format(
                                task
                            )
                        )

                    # These two cases imply faulty TaskGraph construction
                    # Note that this check is only made as this is necessary constraint for decomposing TaskGraph nodes
                    # Faulty construction should be caught at construction of TaskGraph object, including types
                    if len(task_in_edges) != len(task_input_out_edges):
                        raise TypeError(
                            "Decomposition of TaskGraph node {} not permitted: node "
                            "expects {} input wires but receives {}.".format(
                                task, len(task_input_out_edges), len(task_in_edges)
                            )
                        )
                    if len(task_out_edges) != len(task_output_in_edges):
                        raise TypeError(
                            "Decomposition of TaskGraph node {} not permitted: task_graph "
                            "expects {} output wires but node returns {}.".format(
                                task, len(task_output_in_edges), len(task_out_edges)
                            )
                        )

                    # remove all input and output edges from task._task_graph
                    # remove all input and output edges from self._task_graph task
                    # replace in_edges to task in self._task_graph with in_edges to first tasks in task
                    for outside_edge, inside_edge in zip(
                        task_in_edges, task_input_out_edges
                    ):
                        task._task_graph.remove_edge(inside_edge[0], inside_edge[1])
                        self._task_graph.remove_edge(outside_edge[0], outside_edge[1])
                        self._task_graph.add_edge(
                            outside_edge[0],
                            inside_edge[1],
                            key=(outside_edge[2][0], inside_edge[2][1]),
                            data=None,
                        )

                    for outside_edge, inside_edge in zip(
                        task_out_edges, task_output_in_edges
                    ):
                        task._task_graph.remove_edge(inside_edge[0], inside_edge[1])
                        self._task_graph.remove_edge(outside_edge[0], outside_edge[1])
                        self._task_graph.add_edge(
                            inside_edge[0],
                            outside_edge[1],
                            key=(inside_edge[2][0], outside_edge[2][1]),
                            data=None,
                        )

                    # add all remaining edges, filling the rest of the subsituted graph
                    self._task_graph.add_edges_from(task._task_graph.edges)
                    self._task_graph.remove_node(task)
                    check_for_decompose = True
                    break

    def parallel(self, task: Union[MitTask, "TaskGraph"]):
        """
        Adds new MitTask/TaskGraph to TaskGraph object in parallel. All task in edges wired as out edges from Input vertex. All task out_Edges wired as in edges to Output Vertex.

        :param task: New task to be added in parallel.
        """
        task = copy(task)
        base_n_input_outs = len(self._task_graph.out_edges(self._i))
        for port in range(task.n_in_wires):
            self._task_graph.add_edge(
                self._i,
                task,
                key=(base_n_input_outs + port, port),
                data=None,
            )
        base_n_output_ins = len(self._task_graph.in_edges(self._o))
        for port in range(task.n_out_wires):
            self._task_graph.add_edge(
                task,
                self._o,
                key=(port, base_n_output_ins + port),
                data=None,
            )

    def run(
        self, input_wires: List[Wire], cache: bool = False, characterisation: dict = {}
    ) -> Tuple[List[Wire]]:
        """
        Each task in TaskGraph is a pure function that produces output data
        from input data to some specification. Data is stored on edges of the
        internal _task_graph object.
        The run method first assigns each wire in the list to an output edge
        of the input vertex. If more wires are passed than there are edges, wire
        information is not assigned to some edge. If less wires are passed then there are edges,
        some edges will have no data and later computation will likely fail.
        Nodes (holding either MitTask or TaskGraph callable objects) on the graph undergo a topological sort to order them, and are then
        executed sequentially.
        To be executed, data from in edges to a task are passed arguments to the tasks
        _method, and data returned from method are assigned to out edges from a task.
        This process is repeated until all tasks are run, at which point all in edges
        to the output vertex wil have data on, with each data set returned in a tuple.

        :param input_wires: Each Wire holds information assigned as data to an output edge
            from the input vertex of the _task_graph.
        :param cache: If True each Tasks output data is stored in an OrderedDict with the
            Task._label attribute as its key.


        :return: Data from input edges to output vertex, assigned as wires.
        """
        for edge, wire in zip(
            self._task_graph.out_edges(self._i, data=True), input_wires
        ):
            edge[2]["data"] = wire

        # topological_sort fixes any dependency issues so can iterate and assume
        # input wires all realised before a task is reached
        node_list = list(nx.topological_sort(self._task_graph))

        self.characterisation.update(characterisation)
        # clear cache of held data if required
        # also check that all mittask label are unique else dict will fail
        if cache:
            unique_labels = set()
            for task in node_list:
                if task not in (self._i, self._o):
                    unique_labels.add(task._label)
            if len(unique_labels) != len(self._task_graph) - 2:
                raise ValueError(
                    "Cache can't store all information as not all MitTask labels are unique."
                )
            else:
                self._cache.clear()

        for task in node_list:
            # nothing to process
            if task in (self._i, self._o):
                continue
            task.characterisation.update(self.characterisation)
            # get all input data and store on inputs for task
            in_edges = self._task_graph.in_edges(task, data=True, keys=True)
            inputs = [None] * len(in_edges)
            for _, _, ports, i_data in in_edges:
                assert i_data["data"] is not None
                inputs[ports[1]] = i_data["data"]
            # run held task
            outputs = task(inputs)
            self.characterisation.update(task.characterisation)
            if cache:
                self._cache[task._label] = (task, outputs)
            # assign outputs ot out_edges of task
            out_edges = self._task_graph.out_edges(task, data=True, keys=True)
            assert len(out_edges) == len(outputs)
            for _, _, ports, o_data in out_edges:
                o_data["data"] = outputs[ports[0]]

        output_wire = [
            edge[2]["data"]
            for edge in list(self._task_graph.in_edges(self._o, data=True))
        ]
        return cast(Tuple[List[Wire]], tuple(output_wire))

    def get_cache(self) -> OrderedDict[str, Tuple[MitTask, List[Wire]]]:
        """
        :returns: Dictionary holding all output data from all MitTask.
            This is only full after run is called with the cache argument set
            to True. Keys are stored in graph topological order.
        """
        return self._cache

    def get_task_graph(self) -> "gv.Digraph":
        """
        Return a visual representation of the DAG as a graphviz object.

        :returns:   Representation of the DAG
        """
        return _taskgraph_to_graphviz(self._task_graph, None, self._label)

    def view_task_graph(self) -> None:
        """
        View the DAG.
        """
        G = self.get_task_graph()
        file = NamedTemporaryFile()
        G.view(file.name, quiet=True)
